import operator
import re
import tempfile
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Literal
from urllib.parse import parse_qs, urlparse

import requests
import structlog
from platformdirs import user_cache_path

from ytpb.cache import read_from_cache, write_to_cache
from ytpb.config import USER_AGENT
from ytpb.download import compose_default_segment_filename, download_segment
from ytpb.exceptions import (
    BaseUrlExpiredError,
    CachedItemNotFoundError,
    MaxRetryError,
    SequenceLocatingError,
)
from ytpb.fetchers import InfoFetcher, YtpbInfoFetcher
from ytpb.info import LEFT_NOT_FETCHED, LeftNotFetched, YouTubeVideoInfo
from ytpb.locate import SegmentLocator
from ytpb.mpd import extract_representations_info
from ytpb.segment import Segment
from ytpb.streams import SetOfStreams, Streams
from ytpb.types import (
    AbsolutePointInStream,
    PointInStream,
    RelativePointInStream,
    RelativeSegmentSequence,
    SegmentSequence,
)
from ytpb.utils.other import resolve_relativity_in_interval
from ytpb.utils.url import (
    build_video_url_from_base_url,
    check_base_url_is_expired,
    extract_id_from_video_url,
    extract_parameter_from_url,
)

logger = structlog.get_logger(__name__)

SEGMENT_URL_PATTERN = r"https://.+\.googlevideo\.com/videoplayback/.+"


@dataclass(frozen=True)
class RewindMoment:
    date: datetime
    sequence: SegmentSequence
    cut_at: float
    is_end: bool = False
    falls_into_gap: bool = False


@dataclass(frozen=True)
class RewindInterval:
    start: RewindMoment
    end: RewindMoment

    def __post_init__(self):
        if self.start.sequence > self.end.sequence:
            raise ValueError(
                "Start moment is ahead of the end one: "
                f"{self.start.sequence} > {self.end.sequence}"
            )


class PlaybackSession(requests.Session):
    max_retries: int = 3

    def __init__(self, playback: "Playback" = None, **kwargs):
        super().__init__(**kwargs)

        self.playback = playback
        self.hooks["response"].append(self._handle_http_errors)
        self.headers["User-Agent"] = USER_AGENT

    def set_playback(self, playback):
        self.playback = playback

    def _handle_403_error(self, request: requests.Request) -> None:
        old_corresponding_stream = next(
            iter(
                self.playback.streams.filter(
                    lambda x: request.url.startswith(x.base_url)
                )
            )
        )
        old_base_url = old_corresponding_stream.base_url

        self.playback.fetch_and_set_essential()
        new_corresponding_stream = self.playback.streams.get_by_itag(
            old_corresponding_stream.itag
        )
        new_base_url = new_corresponding_stream.base_url

        request.url = request.url.replace(old_base_url, new_base_url)

    def _handle_http_errors(
        self, response: requests.Response, *args, **kwargs
    ) -> requests.Response:
        if response.ok:
            return response

        request = response.request
        retries_count = getattr(request, "retries_count", 0)

        if retries_count < self.max_retries:
            if re.match(SEGMENT_URL_PATTERN, request.url):
                logger.debug("Received %s for %s", response.status_code, request.url)
                logger.debug(
                    "Handle error, and make another try (%s of %s)",
                    retries_count + 1,
                    self.max_retries,
                )
                match response.status_code:
                    case 403:
                        self._handle_403_error(request)
                    case 404:
                        pass
                    case _:
                        logger.debug("Unhandleable error encountered, do nothing")
                        return response
                request.retries_count = retries_count + 1
                return self.send(request, verify=False)
        else:
            raise MaxRetryError(
                f"Maximum number of retries exceeded with URL: {request.url}",
                response,
            )


class Playback:
    def __init__(
        self,
        video_url: str,
        session: requests.Session | None = None,
        fetcher: InfoFetcher | None = None,
        write_to_cache: bool = False,
    ):
        self.video_url = video_url
        self.session = session or PlaybackSession(self)
        self.fetcher = fetcher or YtpbInfoFetcher(video_url)
        self._need_to_cache = write_to_cache

        self._info: YouTubeVideoInfo | LeftNotFetched | None = None
        self._streams: SetOfStreams | None = None
        self._temp_directory: Path | None = None
        self._cache_directory: Path | None = None

    @classmethod
    def from_url(cls, video_url: str, **kwargs) -> "Playback":
        playback = cls(video_url, **kwargs)
        playback.fetch_and_set_essential()
        return playback

    @classmethod
    def from_manifest(
        cls,
        manifest_path: Path,
        fetch_video_info: bool = True,
        **kwargs,
    ) -> "Playback":
        with open(manifest_path, "r") as f:
            list_of_streams = extract_representations_info(f.read())
            streams = Streams(list_of_streams)

        some_base_url = next(iter(streams)).base_url
        video_url = build_video_url_from_base_url(some_base_url)

        if not check_base_url_is_expired(some_base_url):
            playback = cls(video_url, **kwargs)
            playback.set_streams(streams, fetch_video_info=fetch_video_info)
        else:
            logger.debug("Found expired base url, ignore the manifest")
            raise BaseUrlExpiredError

        return playback

    @classmethod
    def from_cache(cls, video_url: str, **kwargs) -> "Playback":
        video_id = extract_id_from_video_url(video_url)
        cached_item = read_from_cache(video_id, Playback.get_cache_directory())
        if cached_item is None:
            raise CachedItemNotFoundError

        playback = cls(video_url, write_to_cache=True, **kwargs)
        playback._info = YouTubeVideoInfo(**cached_item["info"])
        playback._streams = Streams.from_dicts(cached_item["streams"])

        return playback

    @staticmethod
    def get_cache_directory():
        return user_cache_path("ytpb")

    def get_temp_directory(self) -> Path:
        if self._temp_directory is None:
            self._temp_directory = Path(tempfile.mkdtemp(prefix="ytpb-"))
            logger.debug("Run temp directory set to %s", self._temp_directory)
        return self._temp_directory

    @property
    def video_url(self):
        return self._video_url

    @video_url.setter
    def video_url(self, value: str):
        self._video_url = value
        try:
            self._video_id = parse_qs(urlparse(value).query)["v"][0]
        except (KeyError, IndexError):
            print("Could not extract video ID from URL")

    @property
    def video_id(self):
        return self._video_id

    @property
    def info(self) -> YouTubeVideoInfo | LeftNotFetched:
        if self._info is None:
            raise ValueError(
                "Attribute 'info' is not set, call 'fetch_and_set_essential' first"
            )
        return self._info

    @property
    def streams(self) -> SetOfStreams:
        if self._streams is None:
            raise ValueError(
                "Attribute 'streams' is not set, call 'fetch_and_set_essential' first"
            )
        return self._streams

    def _write_to_cache_if_needed(self):
        if self._need_to_cache:
            item_to_cache = {
                "info": asdict(self._info),
                "streams": [asdict(stream) for stream in self.streams],
            }
            some_base_url = next(iter(self.streams)).base_url
            expires_at = extract_parameter_from_url("expire", some_base_url)
            cache_directory = Playback.get_cache_directory()
            write_to_cache(self.video_id, expires_at, item_to_cache, cache_directory)

    def _fetch_and_set_video_info(self) -> None:
        self._info = self.fetcher.fetch_video_info()

    def _fetch_and_set_streams(self) -> None:
        self._streams = self.fetcher.fetch_streams()

    def fetch_and_set_essential(
        self, force_fetch=False, fetch_streams: bool = True
    ) -> None:
        self._fetch_and_set_video_info()
        if fetch_streams:
            self._fetch_and_set_streams()
        self._write_to_cache_if_needed()

    def set_streams(self, value: SetOfStreams, fetch_video_info: bool = True) -> None:
        self._streams = value
        if not self._info or fetch_video_info:
            self._fetch_and_set_video_info()
        else:
            self._info = LEFT_NOT_FETCHED
        self._write_to_cache_if_needed()

    def _get_reference_base_url(self, itag: str) -> str:
        return self.streams.get_by_itag(itag).base_url

    def download_segment(
        self,
        sequence,
        base_url,
        location: Literal[".", "segments"] = ".",
        force_download: bool = False,
    ) -> Path:
        path = download_segment(
            sequence,
            base_url,
            self.get_temp_directory() / location,
            session=self.session,
            force_download=force_download,
        )
        return path

    def get_downloaded_segment(
        self,
        sequence: SegmentSequence,
        base_url: str,
        location: Literal[".", "segments"] = ".",
        download: bool = True,
    ) -> Segment:
        """Get a downloaded segment, or download it if it doesn't exist."""
        segment_filename = compose_default_segment_filename(sequence, base_url)
        segment_directory = self.get_temp_directory() / location
        try:
            segment = Segment.from_file(segment_directory / segment_filename)
        except FileNotFoundError:
            if download:
                downloaded_path = self.download_segment(sequence, base_url)
                segment = Segment.from_file(downloaded_path)
            else:
                raise
        return segment

    def locate_moment(
        self,
        point: AbsolutePointInStream,
        itag: str | None = None,
        is_end: bool = False,
    ) -> SegmentSequence:
        itag = itag or next(iter(self.streams)).itag
        base_url = self._get_reference_base_url(itag)

        match point:
            case SegmentSequence() as sequence:
                self.download_segment(sequence, base_url)
                segment = self.get_downloaded_segment(sequence, base_url)
                if not is_end:
                    date = segment.ingestion_start_date
                else:
                    date = segment.ingestion_end_date
                moment = RewindMoment(date, sequence, 0, is_end)
            case datetime() as date:
                sl = SegmentLocator(
                    base_url,
                    temp_directory=self.get_temp_directory(),
                    session=self.session,
                )
                sequence, falls_into_gap = sl.find_sequence_by_time(
                    point.timestamp(), end=is_end
                )
                segment = self.get_downloaded_segment(sequence, base_url)
                if falls_into_gap:
                    if is_end:
                        date = segment.ingestion_end_date
                    else:
                        date = segment.ingestion_start_date
                    cut_at = 0
                else:
                    cut_at = (date - segment.ingestion_start_date).total_seconds()
                moment = RewindMoment(date, sequence, cut_at, is_end, falls_into_gap)

        return moment

    def locate_interval(
        self,
        start_point: PointInStream,
        end_point: PointInStream,
        itag: str | None = None,
    ) -> RewindInterval:
        # First, resolve relativity where it's possible, without locating:
        start_point, end_point = resolve_relativity_in_interval(start_point, end_point)

        itag = itag or next(iter(self.streams)).itag
        if type(start_point) is type(end_point) is SegmentSequence:
            start_moment = self.locate_moment(start_point, itag=itag)
            end_moment = self.locate_moment(end_point, itag=itag, is_end=True)
            return RewindInterval(start_moment, end_moment)

        # Then, locate end if start is relative:
        if isinstance(start_point, RelativePointInStream):
            end_moment = self.locate_moment(end_point, itag=itag, is_end=True)
        else:
            end_moment = None

        # Finally, iterate over start and end points to locate moments:
        start_and_end_moments: list[RewindMoment | None] = [None, end_moment]
        for index, (point, is_start) in enumerate(
            ((start_point, True), (end_point, False))
        ):
            if isinstance(point, RelativePointInStream):
                start_moment, end_moment = start_and_end_moments
                # Given the current relative point as delta, the resulted point:
                # start/end point = end/start point -/+ delta.
                if is_start:
                    contrary_moment = end_moment
                    contrary_op = operator.sub
                else:
                    contrary_moment = start_moment
                    contrary_op = operator.add

                match point:
                    case RelativeSegmentSequence() as sequence_delta:
                        sequence = contrary_op(contrary_moment.sequence, sequence_delta)
                        moment = self.locate_moment(sequence, itag, not is_start)
                    case timedelta() as time_delta:
                        date = contrary_op(contrary_moment.date, time_delta)
                        moment = self.locate_moment(date, itag, not is_start)
            else:
                moment = self.locate_moment(point, itag, not is_start)

            start_and_end_moments[index] = moment

        try:
            resulted_interval = RewindInterval(*start_and_end_moments)
        except ValueError as exc:
            raise SequenceLocatingError(str(exc)) from exc

        return resulted_interval



# ruff: noqa: E501

import copy
from dataclasses import asdict
from difflib import unified_diff
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, patch

import freezegun

import pytest
from freezegun import freeze_time

from helpers import patched_freezgun_astimezone

from ytpb.info import YouTubeVideoInfo
from ytpb.playback import Playback
from ytpb.streams import Streams
from ytpb.types import AudioOrVideoStream, AudioStream, VideoStream

freezegun.api.FakeDatetime.astimezone = patched_freezgun_astimezone


@pytest.fixture()
def dash_manifest(audio_base_url: str, video_base_url: str) -> str:
    return f"""<?xml version='1.0' encoding='UTF-8'?>
<!--This file is created with ytpb, and expires at 2023-08-16T04:24:48+02:00-->
<MPD xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="urn:mpeg:DASH:schema:MPD:2011" xsi:schemaLocation="urn:mpeg:DASH:schema:MPD:2011 DASH-MPD.xsd" profiles="urn:mpeg:dash:profile:isoff-main:2011" type="static" mediaPresentationDuration="PT66S">
  <Period duration="PT66S">
    <AdaptationSet id="0" mimeType="audio/mp4" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" duration="2000" timescale="1000"/>
      <Representation id="140" codecs="mp4a.40.2" startWithSAP="1" audioSamplingRate="44100">
        <BaseURL>{audio_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
    <AdaptationSet id="1" mimeType="video/webm" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" duration="2000" timescale="1000"/>
      <Representation id="244" codecs="vp9" startWithSAP="1" width="854" height="480" maxPlayoutRate="1" frameRate="30">
        <BaseURL>{video_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
  </Period>
</MPD>
"""


@freeze_time("2023-08-16T02:30:00+02:00", tz_offset=2)
def test_refresh_mpd(
    ytpb_cli_invoke: Callable,
    mock_fetch_and_set_essential: MagicMock,
    streams_in_list: list[AudioOrVideoStream],
    active_live_video_info: YouTubeVideoInfo,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    dash_manifest: str,
    tmp_path: Path,
    expected_out,
) -> None:
    # Given:
    manifest_path = tmp_path / "manifest.mpd"
    with open(manifest_path, "w") as f:
        f.write(dash_manifest)

    # When:
    def mock_essential(obj: Playback, *args, **kwargs):
        updated_streams = copy.deepcopy(streams_in_list)
        for i, stream in enumerate(updated_streams):
            stream_as_dict = asdict(stream)
            stream_as_dict["base_url"] = "https://test/expire/1692153000/"
            if "audio" in stream.mime_type:
                updated_streams[i] = AudioStream(**stream_as_dict)
            else:
                updated_streams[i] = VideoStream(**stream_as_dict)
            obj._streams = Streams(updated_streams)
            obj._info = active_live_video_info

    with patch.object(
        Playback, "fetch_and_set_essential", side_effect=mock_essential, autospec=True
    ):
        result = ytpb_cli_invoke(["mpd", "refresh", str(manifest_path)])

    # Then:
    assert result.exit_code == 0
    assert result.output == expected_out

    with open(manifest_path) as f:
        refreshed_manifest = f.read()

    actual_diff = list(
        unified_diff(
            dash_manifest.splitlines(keepends=True),
            refreshed_manifest.splitlines(keepends=True),
            n=0,
        )
    )
    assert actual_diff == [
        "--- \n",
        "+++ \n",
        "@@ -2 +2 @@\n",
        "-<!--This file is created with ytpb, and expires at 2023-08-16T04:24:48+02:00-->\n",
        "+<!--This file is created with ytpb, and expires at 2023-08-16T04:30:00+02:00-->\n",
        "@@ -8 +8 @@\n",
        f"-        <BaseURL>{audio_base_url}</BaseURL>\n",
        "+        <BaseURL>https://test/expire/1692153000/</BaseURL>\n",
        "@@ -14 +14 @@\n",
        f"-        <BaseURL>{video_base_url}</BaseURL>\n",
        "+        <BaseURL>https://test/expire/1692153000/</BaseURL>\n",
    ]



import sys
from datetime import datetime, timedelta
from pathlib import Path

import click
import cloup
import structlog
from cloup.constraints import constraint, require_any
from rich import box
from rich.console import Console
from rich.table import Table

from ytpb import types
from ytpb.actions.compose import compose_static_mpd, refresh_mpd
from ytpb.cli.commands.download import (
    DownloadOutputPathContext,
    render_download_output_path_context,
)
from ytpb.cli.common import (
    CONSOLE_TEXT_WIDTH,
    create_playback,
    print_summary_info,
    query_streams_or_exit,
    raise_for_sequence_ahead_of_current,
    raise_for_too_far_sequence,
    resolve_output_path,
    stream_argument,
)
from ytpb.cli.options import (
    cache_options,
    interval_option,
    validate_output_path,
    yt_dlp_option,
)
from ytpb.cli.parameters import FormatSpecParamType, FormatSpecType
from ytpb.exceptions import BroadcastStatusError
from ytpb.fetchers import YoutubeDLInfoFetcher, YtpbInfoFetcher
from ytpb.info import BroadcastStatus
from ytpb.mpd import extract_representations_info
from ytpb.playback import Playback
from ytpb.streams import Streams
from ytpb.types import DateInterval, SegmentSequence
from ytpb.utils.date import express_timedelta_in_words
from ytpb.utils.other import resolve_relativity_in_interval
from ytpb.utils.path import expand_template_output_path, OUTPUT_PATH_PLACEHOLDER_RE
from ytpb.utils.remote import request_reference_sequence
from ytpb.utils.url import build_video_url_from_base_url, extract_parameter_from_url

logger = structlog.get_logger(__name__)


MpdOutputPathContext = DownloadOutputPathContext
render_mpd_output_path_context = render_download_output_path_context


def print_audio_table(console, streams, **table_kwargs):
    table = Table(title="(Audio)", **table_kwargs)
    for name in ("itag", "Format", "Codec", "Sampling"):
        if name == "itag":
            table.add_column(" " + name, ratio=0.2, justify="center")
        else:
            table.add_column(name, ratio=(1 - 0.2) / 3)
    for s in sorted(streams, key=lambda x: (x.audio_sampling_rate, x.itag)):
        table.add_row(s.itag, s.format, s.codecs, f"{s.audio_sampling_rate} Hz")
    console.print(table)


def print_video_table(console, streams, **table_kwargs):
    table = Table(title="(Video)", **table_kwargs)
    for name in ("itag", "Format", "Codec", "Quality"):
        if name == "itag":
            table.add_column(" " + name, ratio=0.2, justify="center")
        else:
            table.add_column(name, ratio=(1 - 0.2) / 3)
    for s in sorted(streams, key=lambda x: (x.codecs, x.quality, x.itag)):
        table.add_row(s.itag, s.format, s.codecs, str(s.quality))
    console.print(table)


@cloup.group("mpd", short_help="Compose MPEG-DASH manifests.")
def mpd_group():
    pass


@mpd_group.command(
    "compose",
    help="Compose an MPEG-DASH manifest.",
)
@cloup.option_group(
    "Input options",
    interval_option,
    cloup.option(
        "-af",
        "--audio-formats",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.AUDIO),
        help="Audio format(s) to include.",
    ),
    cloup.option(
        "-vf",
        "--video-formats",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.VIDEO),
        help="Video format(s) to include.",
    ),
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path),
    help="Output path (with extension).",
    callback=validate_output_path(MpdOutputPathContext),
)
@yt_dlp_option
@cache_options
@stream_argument
@constraint(require_any, ["audio_formats", "video_formats"])
@click.pass_context
def compose_command(
    ctx: click.Context,
    interval: types.PointInStream,
    audio_formats: str,
    video_formats: str,
    output_path: Path,
    force_update_cache: bool,
    no_cache: bool,
    yt_dlp: bool,
    stream_url: str,
) -> int:
    playback = create_playback(ctx)

    queried_audio_streams = []
    if audio_formats:
        queried_audio_streams = query_streams_or_exit(
            playback.streams, audio_formats, "--audio-formats"
        )

    queried_video_streams = []
    if video_formats:
        queried_video_streams = query_streams_or_exit(
            playback.streams, video_formats, "--video-formats"
        )

    number_of_streams = len(queried_audio_streams) + len(queried_video_streams)
    if number_of_streams == 1:
        click.echo("This representation will be in the manifest:")
    else:
        click.echo(
            f"These {number_of_streams} representations will be in the manifest:"
        )

    console = Console(width=CONSOLE_TEXT_WIDTH)

    table_kwargs = {
        "box": box.MARKDOWN,
        "padding": (0, 1, 0, 1),
        "show_edge": False,
        "title_style": "italic",
        "header_style": None,
        "expand": True,
        "width": 60,
    }

    if queried_audio_streams:
        print_audio_table(console, queried_audio_streams, **table_kwargs)
    if queried_video_streams:
        print_video_table(console, queried_video_streams, **table_kwargs)
    click.echo()

    reference_stream = (queried_video_streams or queried_audio_streams)[0]
    reference_base_url = reference_stream.base_url
    head_sequence = request_reference_sequence(reference_base_url, playback.session)

    requested_start, requested_end = resolve_relativity_in_interval(*interval)

    if isinstance(requested_start, SegmentSequence):
        raise_for_too_far_sequence(
            requested_start, head_sequence, reference_base_url, ctx, "interval"
        )
        raise_for_sequence_ahead_of_current(
            requested_start, head_sequence, ctx, "interval"
        )

    if isinstance(requested_end, SegmentSequence):
        raise_for_sequence_ahead_of_current(
            requested_end, head_sequence, ctx, "interval"
        )
    if requested_end == "now":
        requested_end = head_sequence - 1

    click.echo("(<<) Locating start and end in the stream... ", nl=False)
    rewind_interval = playback.locate_interval(
        requested_start, requested_end, reference_stream.itag
    )
    click.echo("done.")

    start_segment = playback.get_downloaded_segment(
        rewind_interval.start.sequence, reference_base_url
    )
    end_segment = playback.get_downloaded_segment(
        rewind_interval.end.sequence, reference_base_url
    )

    requested_start_date: datetime
    match requested_start:
        case SegmentSequence():
            requested_start_date = start_segment.ingestion_start_date
        case datetime():
            requested_start_date = requested_start
        case timedelta() as delta:
            requested_start_date = end_segment.ingestion_end_date - delta

    requested_end_date: datetime
    match requested_end:
        case SegmentSequence():
            requested_end_date = end_segment.ingestion_end_date
        case datetime():
            requested_end_date = requested_end
        case timedelta() as delta:
            requested_end_date = start_segment.ingestion_start_date + delta

    requested_date_interval = DateInterval(requested_start_date, requested_end_date)
    actual_date_interval = DateInterval(
        start_segment.ingestion_start_date,
        end_segment.ingestion_end_date,
    )

    print_summary_info(requested_date_interval, actual_date_interval, rewind_interval)
    click.echo()

    # Absolute output path of a manifest with extension.
    final_output_path: Path
    if OUTPUT_PATH_PLACEHOLDER_RE.search(str(output_path)):
        input_timezone = requested_date_interval.start.tzinfo
        template_context: MpdOutputPathContext = {
            "id": playback.video_id,
            "title": playback.info.title,
            "input_start_date": requested_date_interval.start,
            "input_end_date": requested_date_interval.end,
            "actual_start_date": actual_date_interval.start.astimezone(input_timezone),
            "actual_end_date": actual_date_interval.end.astimezone(input_timezone),
            "duration": requested_end_date - requested_start_date,
        }
        final_output_path = expand_template_output_path(
            output_path,
            template_context,
            render_mpd_output_path_context,
            ctx.obj.config,
        )
    else:
        final_output_path = output_path
    final_output_path = resolve_output_path(final_output_path)

    click.echo("(<<) Composing MPEG-DASH manifest...")
    streams = Streams(queried_audio_streams + queried_video_streams)
    composed_manifest = compose_static_mpd(playback, rewind_interval, streams)

    with open(final_output_path, "w") as f:
        f.write(composed_manifest)

    try:
        saved_to_path_value = f"{final_output_path.relative_to(Path.cwd())}"
    except ValueError:
        saved_to_path_value = final_output_path

    click.echo(f"Success! Saved to '{saved_to_path_value}'.")
    click.echo("~ Note that the manifest will expire in 6 hours.")


@mpd_group.command(
    "refresh",
    help="Refresh a composed MPEG-DASH manifest.",
)
@yt_dlp_option
@cloup.argument("manifest", help="Manifest file to refresh.")
def refresh_command(
    yt_dlp: bool,
    manifest: str,
) -> int:
    with open(manifest, "r", encoding="utf-8") as f:
        manifest_content = f.read()

    list_of_streams = extract_representations_info(manifest_content)
    stream_url = build_video_url_from_base_url(list_of_streams[0].base_url)

    if yt_dlp:
        fetcher = YoutubeDLInfoFetcher(stream_url)
    else:
        fetcher = YtpbInfoFetcher(stream_url)

    try:
        playback = Playback.from_url(stream_url, fetcher=fetcher, write_to_cache=False)
    except BroadcastStatusError as e:
        match e.status:
            case BroadcastStatus.COMPLETED:
                click.echo("Stream was live, but now it's finished", err=True)
        sys.exit(1)

    refreshed = refresh_mpd(manifest_content, playback.streams)
    with open(manifest, "w") as f:
        f.write(refreshed)

    some_base_url = next(iter(playback.streams)).base_url
    expires_at_time = int(extract_parameter_from_url("expire", some_base_url))
    expires_at_date = datetime.fromtimestamp(expires_at_time)
    expires_in = express_timedelta_in_words(expires_at_date - datetime.now())

    click.echo(f"Success! Manifest has been refreshed, and expires in {expires_in}.")



import datetime

project = "Ytpb"
copyright = f"{datetime.date.today().year}, Maxim Stolyarchuk"
author = "Maxim Stolyarchuk"

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx_toolbox.collapse",
]

autosectionlabel_prefix_document = True

html_theme = "alabaster"
html_theme_options = {
    "fixed_sidebar": True,
}



"""Provides locating a segment with a desired time."""

import math
import tempfile
from bisect import bisect_left
from functools import partial
from pathlib import Path

import requests
import structlog

from ytpb.download import compose_default_segment_filename, download_segment
from ytpb.exceptions import SegmentDownloadError, SequenceLocatingError
from ytpb.segment import Segment, SegmentMetadata
from ytpb.types import SegmentSequence, Timestamp
from ytpb.utils.remote import request_reference_sequence
from ytpb.utils.url import extract_parameter_from_url

logger = structlog.get_logger(__name__)

# Number of bytes sufficient to cover the YouTube metadata in segment
# files. Note that the minimum value varies for different media formats, so the
# value was determined empirically for all available formats.
PARTIAL_SEGMENT_SIZE_BYTES = 2000

# Absolute time difference tolerance (in seconds). See issue #5.
TIME_DIFF_TOLERANCE = 3e-2


class SequenceMetadataPair:
    def __init__(self, sequence: SegmentSequence, locator: "SegmentLocator"):
        self.locator = locator
        self.sequence = sequence

    @property
    def sequence(self) -> SegmentSequence:
        return self._sequence

    @sequence.setter
    def sequence(self, value: SegmentSequence):
        self._sequence = value
        self._metadata = self._download_segment_and_parse_metadata(value)

    @property
    def metadata(self):
        return self._metadata

    def _download_segment_and_parse_metadata(
        self, sequence: SegmentSequence
    ) -> SegmentMetadata:
        """Download a partial segment and parse metadata."""
        segment_filename = compose_default_segment_filename(
            sequence, self.locator.base_url
        )
        output_filename = segment_filename + ".part"
        downloaded_path = download_segment(
            sequence,
            self.locator.base_url,
            output_directory=self.locator.get_temp_directory(),
            output_filename=output_filename,
            size=PARTIAL_SEGMENT_SIZE_BYTES,
            session=self.locator.session,
            force_download=False,
        )
        with open(downloaded_path, "rb") as f:
            metadata = Segment.parse_youtube_metadata(f.read())
        return metadata


class SegmentLocator:
    """A locator which finds a segment with a desired time.

    A timeline may contain numerous gaps, which leads to under- or
    overestimation, and it needs to be taken into account.

    The locating consists of three steps: (1) a "look around", jump-based search
    to find a segment directly or outline a search domain (the jump length is
    based on the time difference and constant duration of segments); (2) refine
    an estimated sequence number using a binary search if a segment is not
    found; (3) check whether a target time falls into gap or not.
    """

    def __init__(
        self,
        base_url: str,
        reference_sequence: SegmentSequence | None = None,
        temp_directory: Path | None = None,
        session: requests.Session | None = None,
    ) -> None:
        self._temp_directory = temp_directory
        self.base_url = base_url
        self.session = session or requests.Session()

        if reference_sequence is None:
            reference_sequence = request_reference_sequence(base_url)
        self.reference = SequenceMetadataPair(reference_sequence, self)
        self.candidate: SequenceMetadataPair | None = None

        self.segment_duration = float(extract_parameter_from_url("dur", base_url))
        self.track: list[tuple[SegmentSequence, float]] = []

    def get_temp_directory(self):
        if self._temp_directory is None:
            self._temp_directory = tempfile.mkdtemp()
        return self._temp_directory

    def _download_full_segment(self, sequence: SegmentSequence) -> Path:
        downloaded_path = download_segment(
            sequence,
            self.base_url,
            output_directory=self.get_temp_directory(),
            session=self.session,
        )
        return downloaded_path

    def _find_time_diff(
        self, candidate: SequenceMetadataPair, desired_time: Timestamp
    ) -> float:
        return desired_time - candidate.metadata.ingestion_walltime

    def _get_bisected_segment_timestamp(
        self, sequence: SegmentSequence, *, target: Timestamp
    ) -> Timestamp:
        self.candidate.sequence = sequence
        current_diff_in_s = self._find_time_diff(self.candidate, target)
        self.track.append((sequence, current_diff_in_s))
        logger.debug(
            "Bisect step, time difference: %+f s",
            current_diff_in_s,
            seq=self.candidate.sequence,
            time=self.candidate.metadata.ingestion_walltime,
        )
        return self.candidate.metadata.ingestion_walltime

    def _search_sequence(
        self,
        desired_time: Timestamp,
        start: SegmentSequence,
        end: SegmentSequence,
        is_end: bool,
    ) -> tuple[SegmentSequence, bool]:
        """Search for a segment with a desired time inside a search domain with
        the given start and end sequence numbers and perform a gap check."""
        search_domain = range(min((start, end)), max((start, end)) + 1)
        logger.debug("Start a binary search", domain=search_domain)
        bisect_key = partial(self._get_bisected_segment_timestamp, target=desired_time)
        found_index = bisect_left(search_domain, desired_time, key=bisect_key)
        self.candidate.sequence = search_domain[found_index - 1]

        # After the previous step the time difference is always positive.
        candidate_diff_in_s = self._find_time_diff(self.candidate, desired_time)
        if candidate_diff_in_s == 0:
            return self.candidate.sequence, False

        downloaded_path = self._download_full_segment(self.candidate.sequence)
        candidate_segment = Segment.from_file(downloaded_path)
        candidate_duration = candidate_segment.get_actual_duration()

        logger.debug(
            "Candidate time difference: %+f s, actual duration: %f s",
            candidate_diff_in_s,
            candidate_duration,
        )

        falls_into_gap = False
        if candidate_duration < candidate_diff_in_s - TIME_DIFF_TOLERANCE:
            falls_into_gap = True
            logger.debug("Input target time falls into a gap")
            if not is_end:
                self.candidate.sequence += 1
                current_diff_in_s = self._find_time_diff(self.candidate, desired_time)
                self.track.append((self.candidate.sequence, current_diff_in_s))
                logger.debug(
                    "Step to adjacent segment, time difference: %+f s",
                    current_diff_in_s,
                    seq=self.candidate.sequence,
                    time=self.candidate.metadata.ingestion_walltime,
                )

        return self.candidate.sequence, falls_into_gap

    def find_sequence_by_time(
        self, desired_time: Timestamp, end: bool = False
    ) -> tuple[SegmentSequence, bool]:
        """Find sequence number of a segment by the given timestamp."""
        logger.info(
            "Locating segment",
            end=end,
            target=desired_time,
            reference=self.reference.sequence,
        )

        current_diff_in_s = self._find_time_diff(self.reference, desired_time)
        start_direction = int(math.copysign(1, current_diff_in_s))
        self.candidate = self.reference

        has_segment_found = False

        has_domain_discovered = False
        has_sign_changed = False

        while not has_domain_discovered:
            self.track.append((self.candidate.sequence, current_diff_in_s))
            logger.debug(
                "Look around, time difference: %+f s",
                current_diff_in_s,
                seq=self.candidate.sequence,
                time=self.candidate.metadata.ingestion_walltime,
            )

            if 0 <= current_diff_in_s <= self.segment_duration + TIME_DIFF_TOLERANCE:
                has_segment_found = True
                break

            direction = int(math.copysign(1, current_diff_in_s))
            if not has_sign_changed:
                has_sign_changed = direction * start_direction < 0
            has_domain_discovered = has_sign_changed and direction == start_direction

            jump_length_in_seq = int(current_diff_in_s // self.segment_duration)
            self.candidate = SequenceMetadataPair(
                self.candidate.sequence + jump_length_in_seq, self
            )
            current_diff_in_s = self._find_time_diff(self.candidate, desired_time)

        result: tuple[SegmentSequence, bool]
        if has_segment_found:
            result = (self.candidate.sequence, False)
        else:
            try:
                search_between = (self.track[-2][0], self.track[-1][0])
                result = self._search_sequence(desired_time, *search_between, end)
            except SegmentDownloadError as exc:
                raise SequenceLocatingError(exc) from exc

        logger.info("Segment has been located as %d", self.candidate.sequence)

        return result



import shutil
import sys
from datetime import datetime, timedelta
from functools import partial
from pathlib import Path

import click
import cloup
import structlog
from cloup.constraints import constraint, require_any

from ytpb import actions
from ytpb.cli.common import (
    create_playback,
    get_parameter_by_name,
    print_summary_info,
    query_streams_or_exit,
    raise_for_sequence_ahead_of_current,
    raise_for_too_far_sequence,
    resolve_output_path,
    stream_argument,
)
from ytpb.cli.options import (
    cache_options,
    interval_option,
    no_cleanup_option,
    preview_option,
    validate_output_path,
    yt_dlp_option,
)
from ytpb.cli.parameters import FormatSpecParamType, FormatSpecType, InputRewindInterval
from ytpb.exceptions import SequenceLocatingError
from ytpb.merge import merge_segments
from ytpb.types import (
    AddressableMappingProtocol,
    DateInterval,
    RelativeSegmentSequence,
    SegmentSequence,
)
from ytpb.utils.other import resolve_relativity_in_interval, S_TO_MS
from ytpb.utils.path import (
    expand_template_output_path,
    IntervalOutputPathContext,
    MinimalOutputPathContext,
    OUTPUT_PATH_PLACEHOLDER_RE,
    render_interval_output_path_context,
    render_minimal_output_path_context,
)
from ytpb.utils.remote import request_reference_sequence
from ytpb.utils.url import extract_parameter_from_url

logger = structlog.get_logger(__name__)


class DownloadOutputPathContext(
    MinimalOutputPathContext, IntervalOutputPathContext
): ...


def render_download_output_path_context(
    context: DownloadOutputPathContext,
    config_settings: AddressableMappingProtocol,
) -> DownloadOutputPathContext:
    output = context
    output.update(render_minimal_output_path_context(context, config_settings))
    output.update(render_interval_output_path_context(context, config_settings))
    return output


@cloup.command("download", short_help="Download excerpts.", help="Download an excerpt.")
@cloup.option_group(
    "Input options",
    interval_option,
    cloup.option(
        "-af",
        "--audio-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.AUDIO),
        help="Audio format to download.",
    ),
    cloup.option(
        "-vf",
        "--video-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.VIDEO),
        help="Video format to download.",
    ),
    preview_option,
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path),
    help="Output path (without extension).",
    callback=validate_output_path(DownloadOutputPathContext),
)
@click.option(
    "-m",
    "--from-manifest",
    metavar="PATH",
    type=click.Path(path_type=Path),
    help="Path to a MPEG-DASH manifest.",
)
@click.option(
    "-X",
    "--dry-run",
    is_flag=True,
    help=("Run without downloading."),
)
@yt_dlp_option
@click.option("--no-cut", is_flag=True, help="Do not perform excerpt cutting.")
@click.option(
    "--no-merge",
    is_flag=True,
    help=("Only download segments, without merging. This implies '--no-cleanup'."),
)
@no_cleanup_option
@cache_options
@stream_argument
@constraint(require_any, ["audio_format", "video_format"])
@click.pass_context
def download_command(
    ctx: click.Context,
    interval: InputRewindInterval,
    audio_format: str,
    video_format: str,
    preview: bool,
    output_path: Path,
    from_manifest: Path,
    dry_run: bool,
    yt_dlp: bool,
    no_cut: bool,
    no_merge: bool,
    no_cleanup: bool,
    force_update_cache: bool,
    no_cache: bool,
    stream_url: str,
) -> int:
    if no_merge:
        no_cleanup = True

    playback = create_playback(ctx)

    queried_audio_streams = []
    if audio_format:
        logger.debug("Query audio stream by format spec", spec=audio_format)
        queried_audio_streams = query_streams_or_exit(
            playback.streams, audio_format, "--audio-format", allow_many=False
        )

    queried_video_streams = []
    if video_format:
        logger.debug("Query video stream by format spec", spec=video_format)
        queried_video_streams = query_streams_or_exit(
            playback.streams, video_format, "--video-format", allow_many=False
        )

    if queried_audio_streams and queried_video_streams:
        click.echo("These representations will be downloaded:")
    else:
        click.echo("This representation will be downloaded:")

    audio_stream, video_stream = None, None
    if audio_format:
        (audio_stream,) = queried_audio_streams
        click.echo(
            f"   - Audio: itag {audio_stream.itag}, "
            f"{audio_stream.format} ({audio_stream.codecs}), "
            f"{audio_stream.audio_sampling_rate} Hz"
        )
        logger.info("Queried audio stream", base_url=audio_stream.base_url)

    if video_format:
        (video_stream,) = queried_video_streams
        click.echo(
            f"   - Video: itag {video_stream.itag}, "
            f"{video_stream.format} ({video_stream.codecs}), "
            f"{video_stream.width}x{video_stream.height}, "
            f"{video_stream.frame_rate} fps"
        )
        logger.info("Queried video stream", base_url=video_stream.base_url)

    click.echo()
    click.echo("(<<) Locating start and end in the stream... ", nl=False)

    reference_stream = video_stream or audio_stream
    reference_base_url = reference_stream.base_url
    head_sequence = request_reference_sequence(reference_base_url, playback.session)

    requested_start, requested_end = resolve_relativity_in_interval(*interval)

    if isinstance(requested_start, SegmentSequence):
        raise_for_too_far_sequence(
            requested_start, head_sequence, reference_base_url, ctx, "interval"
        )
        raise_for_sequence_ahead_of_current(
            requested_start, head_sequence, ctx, "interval"
        )

    match requested_end:
        case SegmentSequence() as x:
            raise_for_sequence_ahead_of_current(x, head_sequence, ctx, "interval")
        case "now":
            requested_end = head_sequence - 1
        case "..":
            if not preview:
                raise click.BadParameter(
                    "Open interval is only valid in the preview mode",
                    param=get_parameter_by_name("interval", ctx),
                )

    if preview:
        preview_duration_value = ctx.obj.config.traverse("general.preview_duration")
        segment_duration = float(extract_parameter_from_url("dur", reference_base_url))
        number_of_segments = round(preview_duration_value / segment_duration)
        requested_end = RelativeSegmentSequence(number_of_segments)

    try:
        rewind_interval = playback.locate_interval(
            requested_start,
            requested_end,
            reference_stream.itag,
        )
    except SequenceLocatingError:
        message = "\nerror: An error occured during segment locating, exit."
        click.echo(message, err=True)
        sys.exit(1)

    click.echo("done.")

    if preview and interval[1] != "..":
        click.echo("info: The preview mode is enabled, interval end is ignored.")

    start_segment = playback.get_downloaded_segment(
        rewind_interval.start.sequence, reference_base_url
    )
    end_segment = playback.get_downloaded_segment(
        rewind_interval.end.sequence, reference_base_url
    )

    requested_start_date: datetime
    match requested_start:
        case SegmentSequence():
            requested_start_date = start_segment.ingestion_start_date
        case datetime():
            requested_start_date = requested_start
        case timedelta() as delta:
            requested_start_date = end_segment.ingestion_end_date - delta

    requested_end_date: datetime
    match requested_end:
        case SegmentSequence():
            requested_end_date = end_segment.ingestion_end_date
        case datetime():
            requested_end_date = requested_end
        case timedelta() as delta:
            requested_end_date = start_segment.ingestion_start_date + delta

    requested_date_interval = DateInterval(requested_start_date, requested_end_date)
    actual_date_interval = DateInterval(
        start_segment.ingestion_start_date,
        end_segment.ingestion_end_date,
    )

    if no_cut:
        cut_kwargs = {}
    else:
        start_diff, end_diff = requested_date_interval - actual_date_interval
        cut_at_start_s = start_diff if start_diff > 0 else 0
        if preview:
            cut_at_end_s = 0
        else:
            cut_at_end_s = abs(end_diff) if end_diff < 0 else 0
        cut_kwargs = {
            "cut_at_start": round(cut_at_start_s * int(S_TO_MS)),
            "cut_at_end": round(cut_at_end_s * int(S_TO_MS)),
        }
        if not no_merge:
            actual_date_interval = DateInterval(
                actual_date_interval.start + timedelta(seconds=cut_at_start_s),
                actual_date_interval.end - timedelta(seconds=cut_at_end_s),
            )

    print_summary_info(requested_date_interval, actual_date_interval, rewind_interval)
    click.echo()

    if dry_run:
        click.echo("notice: This is a dry run. Skip downloading and exit.")
    else:
        # Absolute output path of an excerpt without extension.
        final_output_path: Path

        if OUTPUT_PATH_PLACEHOLDER_RE.search(str(output_path)):
            input_timezone = requested_date_interval.start.tzinfo
            template_context: DownloadOutputPathContext = {
                "id": playback.video_id,
                "title": playback.info.title,
                "input_start_date": requested_date_interval.start,
                "input_end_date": requested_date_interval.end,
                "actual_start_date": actual_date_interval.start.astimezone(
                    input_timezone
                ),
                "actual_end_date": actual_date_interval.end.astimezone(input_timezone),
                "duration": requested_end_date - requested_start_date,
            }
            final_output_path = expand_template_output_path(
                output_path,
                template_context,
                render_download_output_path_context,
                ctx.obj.config,
            )
        else:
            final_output_path = output_path
        final_output_path = resolve_output_path(final_output_path)

        do_download_excerpt_segments = partial(
            actions.download.download_excerpt,
            playback,
            rewind_interval,
            audio_format,
            video_format,
            output_directory=final_output_path.parent,
            output_stem=final_output_path.name,
            no_merge=True,
        )

        if no_merge:
            click.echo(
                "(<<) Downloading segments {}-{} (no merge requested)...".format(
                    rewind_interval.start.sequence, rewind_interval.end.sequence
                )
            )
            download_result = do_download_excerpt_segments()
            some_downloaded_paths = (
                download_result.audio_segment_paths
                or download_result.video_segment_paths
            )
            click.echo(
                "\nSuccess! Segments saved to {}/.".format(
                    some_downloaded_paths[0].parent
                )
            )
        else:
            click.echo("(<<) Preparing and saving the excerpt...")
            click.echo(
                "1. Downloading segments {}-{}:".format(
                    rewind_interval.start.sequence, rewind_interval.end.sequence
                )
            )

            download_result = do_download_excerpt_segments()

            if no_cut:
                click.echo("2. Merging segments (no cut requested)... ", nl=False)
            else:
                click.echo("2. Merging segments (may take a while)... ", nl=False)

            merged_path = merge_segments(
                download_result.audio_segment_paths,
                download_result.video_segment_paths,
                output_directory=final_output_path.parent,
                output_stem=final_output_path.name,
                temp_directory=playback.get_temp_directory(),
                **cut_kwargs,
            )
            click.echo("done.\n")

            try:
                saved_to_path_value = f"{merged_path.relative_to(Path.cwd())}"
            except ValueError:
                saved_to_path_value = merged_path
            click.echo(f"Success! Saved to '{saved_to_path_value}'.")

    run_temp_directory = playback.get_temp_directory()
    if no_cleanup:
        click.echo(f"notice: No cleanup enabled, check {run_temp_directory}/")
    else:
        try:
            shutil.rmtree(run_temp_directory)
        except OSError:
            logger.warning(
                "Failed to remove %s temporary directory", run_temp_directory
            )



import pytest
import responses

from ytpb.exceptions import YtpbError
from ytpb.utils.remote import request_reference_sequence


def test_request_reference_sequence(
    mocked_responses: responses.RequestsMock,
    audio_base_url: str,
) -> None:
    mocked_responses.head(audio_base_url, headers={"X-Head-Seqnum": "123"})
    assert request_reference_sequence(audio_base_url) == 123


def test_request_reference_sequence_with_missing_header_value(
    mocked_responses: responses.RequestsMock,
    audio_base_url: str,
) -> None:
    mocked_responses.head(audio_base_url, headers={})

    with pytest.raises(YtpbError):
        request_reference_sequence(audio_base_url)



import json
import os
import re
import tempfile
from dataclasses import asdict
from pathlib import Path
from typing import Callable
from unittest.mock import Mock, patch

import pytest
import responses
from pytest_socket import disable_socket

from ytpb.info import BroadcastStatus, YouTubeVideoInfo
from ytpb.playback import InfoFetcher, Playback
from ytpb.segment import Segment
from ytpb.streams import Streams
from ytpb.types import AudioOrVideoStream, AudioStream, VideoStream
from ytpb.utils.url import extract_parameter_from_url


TEST_DATA_PATH = Path(__file__).parent / "data"


def pytest_runtest_setup():
    disable_socket()


@pytest.fixture(scope="class")
def monkeyclass():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture
def mocked_responses() -> responses.RequestsMock:
    with responses.RequestsMock() as responses_mock:
        yield responses_mock


@pytest.fixture(autouse=True)
def set_enviromental_variables(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, run_temp_directory: Path
):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
    monkeypatch.setenv("XDG_CACHE_HOME", str(tmp_path / "cache"))
    monkeypatch.setattr(
        "ytpb.playback.Playback.get_temp_directory",
        Mock(return_value=run_temp_directory),
    )


@pytest.fixture()
def cache_directory() -> Path:
    cache_directory_path = Path(os.environ["XDG_CACHE_HOME"])
    cache_directory_path.mkdir(parents=True)
    return cache_directory_path


@pytest.fixture()
def create_cache_file(
    monkeypatch: pytest.MonkeyPatch,
    video_id: str,
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    tmp_path: Path,
) -> None:
    test_cache_directory = Path(os.environ["XDG_CACHE_HOME"]) / "ytpb"
    test_cache_directory.mkdir(parents=True)

    some_base_url = streams_in_list[0].base_url
    expired_at = int(some_base_url.split("/expire/")[1].split("/")[0])

    with open(test_cache_directory / f"{expired_at}~{video_id}", "w") as f:
        test_cached_item = {
            "info": asdict(active_live_video_info),
            "streams": [asdict(stream) for stream in streams_in_list],
        }
        json.dump(test_cached_item, f)


@pytest.fixture()
def run_temp_directory(tmp_path: Path) -> Path:
    return Path(tempfile.mkdtemp(dir=tmp_path))


@pytest.fixture()
def video_id() -> str:
    return "kHwmzef842g"


@pytest.fixture()
def stream_url(video_id: str) -> str:
    return f"https://www.youtube.com/watch?v={video_id}"


@pytest.fixture()
def short_stream_url(video_id: str) -> str:
    return f"https://youtu.be/{video_id}"


@pytest.fixture()
def fake_info_fetcher(
    stream_url: str, active_live_video_info, streams_in_list
) -> "FakeInfoFetcher":
    class FakeInfoFetcher(InfoFetcher):
        def fetch_video_info(self):
            return active_live_video_info

        def fetch_streams(self):
            return Streams(streams_in_list)

    return FakeInfoFetcher(stream_url)


@pytest.fixture()
def mock_playback_info_fetcher(fake_info_fetcher):
    with patch("ytpb.playback.YtpbInfoFetcher") as mock:
        mock.return_value = fake_info_fetcher
        yield mock


@pytest.fixture()
def mock_fetch_and_set_essential(
    streams_in_list: list[AudioOrVideoStream], active_live_video_info: YouTubeVideoInfo
):
    def _mocked_fetch_and_set_essential(self: Playback, *args, **kwargs):
        self._streams = Streams(streams_in_list)
        self._info = active_live_video_info

    with patch.object(
        Playback, "fetch_and_set_essential", _mocked_fetch_and_set_essential
    ) as mock:
        yield mock


def load_segment_from_file_callback(location: str, filename: str | None = None):
    def request_callback(request):
        sequence = extract_parameter_from_url("sq", request.url)

        if filename:
            filepath = Path(location) / filename
        else:
            match extract_parameter_from_url("mime", request.url).split("%2F"):
                case ("video", "webm"):
                    filepath = Path(location) / f"{sequence}.i244.webm"
                case ("audio", "mp4"):
                    filepath = Path(location) / f"{sequence}.i140.mp4"
                case _:
                    raise ValueError("No segments matching the request URL")

        try:
            with open(filepath, "rb") as f:
                if range_value := request.headers.get("Range"):
                    size = int(range_value.split("-")[-1])
                else:
                    size = None
                segment = f.read(size)
        except FileNotFoundError as e:
            pytest.exit(e.strerror)

        return 200, {}, segment

    return request_callback


@pytest.fixture()
def add_responses_callback_for_segment_urls(mocked_responses) -> Callable:
    def wrapper(*matching_url_or_urls: str) -> None:
        matching_url_re = re.compile("|".join(matching_url_or_urls))
        mocked_responses.add_callback(
            responses.GET,
            matching_url_re,
            load_segment_from_file_callback(TEST_DATA_PATH / "segments"),
        )

    return wrapper


@pytest.fixture()
def add_responses_callback_for_reference_base_url(mocked_responses) -> Callable:
    def wrapper(reference_sequence: int = 7959203):
        mocked_responses.add_callback(
            responses.HEAD,
            re.compile(r"https://.+\.googlevideo\.com/videoplayback/.+"),
            lambda _: (200, {"X-Head-Seqnum": str(reference_sequence)}, ""),
        )

    return wrapper


@pytest.fixture()
def active_live_video_info(youtube_dl_info: dict) -> YouTubeVideoInfo:
    return YouTubeVideoInfo(
        url=youtube_dl_info["webpage_url"],
        title=youtube_dl_info["title"],
        author=youtube_dl_info["uploader"],
        status=BroadcastStatus.ACTIVE,
        dash_manifest_url=youtube_dl_info["formats"][0]["manifest_url"],
    )


@pytest.fixture()
def audio_base_url(youtube_dl_info) -> str:
    return "https://rr5---sn-25ge7nzr.googlevideo.com/videoplayback/expire/1695928670/ei/_nwVZYXhAqbQvdIPjKmqgAM/ip/0.0.0.0/id/kHwmzef842g.2/itag/140/source/yt_live_broadcast/requiressl/yes/spc/UWF9fy2D4rPPhPMeyQnmxgP0Yhyaohs/vprv/1/playlist_type/DVR/ratebypass/yes/mime/audio%2Fmp4/live/1/gir/yes/noclen/1/dur/2.000/keepalive/yes/fexp/24007246/beids/24350017/sparams/expire,ei,ip,id,itag,source,requiressl,spc,vprv,playlist_type,ratebypass,mime,live,gir,noclen,dur/sig/AOq0QJ8wRAIgANge9FK8aJnP8nDX_HCd9LixBc1iiZueVKgR1eWAi4ACIE5wyoXt2JUnPjHbh6xp8ZJhy1j9ScEgHiBAO_2xH3h9/initcwndbps/623750/mh/XB/mm/44/mn/sn-25ge7nzr/ms/lva/mt/1695906793/mv/m/mvi/5/pl/38/lsparams/initcwndbps,mh,mm,mn,ms,mv,mvi,pl/lsig/AG3C_xAwRQIhAM6lQ9DNT724pGLtqWR01mXgxu_67Ing2nzPBj4ffCT8AiAYnVuWcAosv-DKUGO2bNSq5ptYGJhRCdlYo8E3-6HKOA%3D%3D/"


@pytest.fixture()
def video_base_url() -> str:
    return "https://rr5---sn-25ge7nzr.googlevideo.com/videoplayback/expire/1695928670/ei/_nwVZYXhAqbQvdIPjKmqgAM/ip/0.0.0.0/id/kHwmzef842g.2/itag/244/source/yt_live_broadcast/requiressl/yes/spc/UWF9fy2D4rPPhPMeyQnmxgP0Yhyaohs/vprv/1/playlist_type/DVR/ratebypass/yes/mime/video%2Fwebm/live/1/gir/yes/noclen/1/dur/2.000/keepalive/yes/fexp/24007246/beids/24350017/sparams/expire,ei,ip,id,itag,source,requiressl,spc,vprv,playlist_type,ratebypass,mime,live,gir,noclen,dur/sig/AOq0QJ8wRgIhAJBYRElUjO7WhY5_gsjtj0aUbXbyb9Z_Yjo7JeecnqrzAiEAkzwV4SYIFponf7BddjJ5hscSZr8hbPBSx09Qffev9AA%3D/initcwndbps/623750/mh/XB/mm/44/mn/sn-25ge7nzr/ms/lva/mt/1695906793/mv/m/mvi/5/pl/38/lsparams/initcwndbps,mh,mm,mn,ms,mv,mvi,pl/lsig/AG3C_xAwRQIhAP_FmY_xO0cSx-hk2oibYFE1AHaCvDHeYyMXXUEuBNeVAiARmaf6MprHE-eEJJx3Ai59WyTOSt8INUUWhA7MSoEO2w%3D%3D/"


@pytest.fixture()
def youtube_dl_info() -> dict:
    with open(TEST_DATA_PATH / "info-1695928670.json") as f:
        return json.load(f)


@pytest.fixture()
def streams_in_list(youtube_dl_info: dict) -> list[AudioOrVideoStream]:
    list_of_streams: list[AudioOrVideoStream] = []
    for f in youtube_dl_info["formats"]:
        base_url = f["fragment_base_url"]
        raw_mime_type = extract_parameter_from_url("mime", base_url)
        mime_type = raw_mime_type.replace("%2F", "/")

        attributes = {
            "itag": f["format_id"],
            "base_url": base_url,
            "mime_type": mime_type,
        }
        if f["acodec"] != "none":
            attributes.update(
                {
                    "codecs": f["acodec"],
                    "audio_sampling_rate": f["asr"],
                }
            )
            stream = AudioStream(**attributes)
        else:
            attributes.update(
                {
                    "codecs": f["vcodec"],
                    "width": f["width"],
                    "height": f["height"],
                    "frame_rate": f["fps"],
                }
            )
            stream = VideoStream(**attributes)
        list_of_streams.append(stream)

    return list_of_streams


@pytest.fixture()
def audio_segment() -> Segment:
    yield Segment.from_file(TEST_DATA_PATH / "segments" / "7959120.i140.mp4")


@pytest.fixture()
def next_audio_segment() -> Segment:
    yield Segment.from_file(TEST_DATA_PATH / "segments" / "7959121.i140.mp4")



""" Fixtures for the CLI tests."""

import contextlib
import os
from collections.abc import Callable
from functools import partial
from pathlib import Path

import pytest
from click.testing import CliRunner

from ytpb.cli import cli


@contextlib.contextmanager
def isolated_filesystem(path: Path):
    old_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@pytest.fixture()
def ytpb_cli_invoke(tmp_path: Path) -> Callable:
    runner = CliRunner()
    with isolated_filesystem(tmp_path):
        yield partial(runner.invoke, cli)



from abc import ABC, abstractmethod

import requests
import structlog
from yt_dlp import DownloadError, YoutubeDL

from ytpb.exceptions import BroadcastStatusError
from ytpb.info import BroadcastStatus, extract_video_info, YouTubeVideoInfo
from ytpb.mpd import extract_representations_info
from ytpb.streams import Streams
from ytpb.types import AudioOrVideoStream, AudioStream, VideoStream
from ytpb.utils.url import extract_parameter_from_url

logger = structlog.get_logger(__name__)


class InfoFetcher(ABC):
    def __init__(self, video_url: str, session: requests.Session | None = None):
        self.video_url = video_url
        self.session = session or requests.Session()

    @abstractmethod
    def fetch_video_info(self):
        raise NotImplementedError

    @abstractmethod
    def fetch_streams(self, force_fetch: bool = True):
        raise NotImplementedError


class YtpbInfoFetcher(InfoFetcher):
    def fetch_video_info(self):
        logger.debug("Fetching index webpage and extracting video info")

        response = self.session.get(self.video_url)
        response.raise_for_status()

        info = extract_video_info(self.video_url, response.text)
        if info.status != BroadcastStatus.ACTIVE:
            raise BroadcastStatusError("Stream is not live", info.status)
        self.info = info

        return self.info

    def fetch_streams(self, force_fetch: bool = True):
        logger.debug("Fetching manifest file and extracting streams info")

        dash_manifest_url = self.info.dash_manifest_url
        response = self.session.get(dash_manifest_url)
        response.raise_for_status()

        streams_list = extract_representations_info(response.text)
        streams = Streams(streams_list)

        return streams


class YoutubeDLInfoFetcher(InfoFetcher):
    options = {
        "live_from_start": True,
        "quiet": True,
    }

    def __init__(self, video_url: str, session: requests.Session | None = None):
        super().__init__(video_url, session)
        self._ydl = YoutubeDL(self.options)
        self._formats: list[dict] = []

    def fetch_video_info(self):
        try:
            extracted = self._ydl.extract_info(self.video_url, download=False)
        except DownloadError as exc:
            raise AssertionError from exc

        try:
            self._formats = extracted["formats"]

            match extracted["live_status"]:
                case "is_live":
                    status = BroadcastStatus.ACTIVE
                case "was_live" | "post_live":
                    status = BroadcastStatus.COMPLETED
                case "is_upcoming":
                    status = BroadcastStatus.UPCOMING
                case _:
                    status = BroadcastStatus.NONE
            if status != BroadcastStatus.ACTIVE:
                raise BroadcastStatusError("Stream is not live", status)

            info = YouTubeVideoInfo(
                url=extracted["webpage_url"],
                title=extracted["title"],
                author=extracted["uploader"],
                status=status,
                dash_manifest_url=extracted["formats"][0]["manifest_url"],
            )
        except KeyError:
            raise KeyError("Failed to parse extracted info")

        return info

    def _parse_format_item(self, item: dict) -> AudioOrVideoStream:
        base_url = item["fragment_base_url"]
        raw_mime_type = extract_parameter_from_url("mime", base_url)
        mime_type = raw_mime_type.replace("%2F", "/")

        attributes = {
            "itag": item["format_id"],
            "base_url": base_url,
            "mime_type": mime_type,
        }
        if item["acodec"] != "none":
            attributes.update(
                {
                    "codecs": item["acodec"],
                    "audio_sampling_rate": item["asr"],
                }
            )
            stream = AudioStream(**attributes)
        else:
            attributes.update(
                {
                    "codecs": item["vcodec"],
                    "width": item["width"],
                    "height": item["height"],
                    "frame_rate": item["fps"],
                }
            )
            stream = VideoStream(**attributes)
        return stream

    def fetch_streams(self, force_fetch: bool = True):
        streams = Streams()

        if not self._formats or force_fetch:
            self.fetch_video_info()

        for format_item in self._formats:
            try:
                stream = self._parse_format_item(format_item)
                streams.add(stream)
            except KeyError:
                raise AssertionError

        return streams



import re
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path

import av
import structlog

from ytpb.exceptions import YtpbError
from ytpb.types import SegmentSequence, Timestamp
from ytpb.utils.other import US_TO_S

logger = structlog.get_logger(__name__)


@dataclass
class SegmentMetadata:
    sequence_number: SegmentSequence
    ingestion_walltime: Timestamp
    ingestion_uncertainty: float
    stream_duration: float
    max_dvr_duration: float
    target_duration: float
    first_frame_time: Timestamp
    first_frame_uncertainty: float
    streamable: str | None = None
    encoding_alias: str | None = None


@dataclass
class Segment:
    def __init__(self) -> None:
        self.local_path: Path | None = None
        self.metadata: SegmentMetadata | None = None
        self.sequence: SegmentSequence | None = None
        self.is_partial: bool | None = None

    @classmethod
    def from_file(cls, path: Path) -> "Segment":
        segment = cls()

        with open(path, "rb") as f:
            content = f.read()
            segment.local_path = path

        segment.metadata = Segment.parse_youtube_metadata(content)
        segment.sequence = segment.metadata.sequence_number

        return segment

    @classmethod
    def from_bytes(cls, content: bytes) -> "Segment":
        segment = cls()
        segment.metadata = Segment.parse_youtube_metadata(content)
        segment.sequence = segment.metadata.sequence_number
        segment.is_partial = True

        return segment

    @property
    def ingestion_start_date(self):
        timestamp = self.metadata.ingestion_walltime
        return datetime.fromtimestamp(timestamp, timezone.utc)

    @property
    def ingestion_end_date(self):
        return self.ingestion_start_date + timedelta(seconds=self.get_actual_duration())

    @staticmethod
    def parse_youtube_metadata(content: bytes) -> SegmentMetadata:
        optional_fields = ("Streamable", "Encoding-Alias")

        def _search_for_metadata_field(
            name: str, content: bytes, optional: bool = False
        ) -> bytes | None:
            if matched := re.search(rf"{name}:\s(.+)\r\n".encode(), content):
                value = matched.group(1)
            else:
                if not optional:
                    raise YtpbError(f"Failed to parse metadata field: {name}")
                value = None
            return value

        def _convert_to_float_in_s(value: bytes) -> float:
            return float(value.decode()) / (1 / US_TO_S)

        def _convert_to_timestamp_in_s(value: bytes) -> Timestamp:
            return _convert_to_float_in_s(value)

        metadata_fields_map = (
            ("Sequence-Number", lambda x: int(x.decode())),
            ("Ingestion-Walltime-Us", _convert_to_timestamp_in_s),
            ("Ingestion-Uncertainty-Us", _convert_to_float_in_s),
            ("Stream-Duration-Us", _convert_to_float_in_s),
            ("Max-Dvr-Duration-Us", _convert_to_float_in_s),
            ("Target-Duration-Us", _convert_to_float_in_s),
            ("Streamable", lambda x: x.decode()),
            ("First-Frame-Time-Us", _convert_to_timestamp_in_s),
            ("First-Frame-Uncertainty-Us", _convert_to_float_in_s),
            ("Encoding-Alias", lambda x: x.decode()),
        )

        parsed_metadata_fields = {}
        for name, cast_func in metadata_fields_map:
            value_bytes = _search_for_metadata_field(
                name, content, optional=name in optional_fields
            )
            if value_bytes:
                value = cast_func(value_bytes)
                name_as_key = name.removesuffix("-Us").lower().replace("-", "_")
                parsed_metadata_fields[name_as_key] = value

        return SegmentMetadata(**parsed_metadata_fields)

    def get_actual_duration(self) -> float:
        with av.open(str(self.local_path)) as container:
            first_packet, *_, last_packet = list(container.demux())[:-1]
            end_timestamp = last_packet.pts + last_packet.duration
            duration = (end_timestamp - first_packet.pts) * float(
                first_packet.time_base
            )
            return duration



import pytest

from ytpb.exceptions import QueryError
from ytpb.streams import AudioOrVideoStream, AudioStream, Streams


def test_abstract_methods(streams_in_list: list[AudioOrVideoStream]):
    streams = Streams(streams_in_list)

    assert len(streams) == len(streams_in_list)
    assert isinstance(next(iter(streams)), AudioOrVideoStream)

    unknown_stream = AudioStream(
        itag="0",
        mime_type="test",
        codecs="test",
        audio_sampling_rate=0,
        base_url="test",
    )
    assert streams_in_list[0] in streams
    assert unknown_stream not in streams


def test_filter(streams_in_list: list[AudioOrVideoStream]):
    streams = Streams(streams_in_list)
    filtered = streams.filter(lambda x: isinstance(x, AudioStream))
    assert len(filtered) == 2
    assert isinstance(next(iter(filtered)), AudioStream)


def test_non_empty_query(streams_in_list: list[AudioOrVideoStream]):
    streams = Streams(streams_in_list)
    assert streams.query("itag eq 244") == [streams.get_by_itag("244")]


def test_empty_query(streams_in_list: list[AudioOrVideoStream]):
    assert Streams(streams_in_list).query("itag eq 0") == []


@pytest.mark.parametrize(
    "format_spec",
    [
        "function(cond1",
        "function(cond1) and cond2",
        "function()",
        "(cond1 and cond2)",
    ],
)
def test_invalid_query(format_spec: str, streams_in_list: list[AudioOrVideoStream]):
    with pytest.raises(QueryError) as exc_info:
        Streams(streams_in_list).query(format_spec)
    assert "Format spec is invalid" in str(exc_info.value)


def test_query_with_alias(streams_in_list: list[AudioOrVideoStream]):
    streams = Streams(streams_in_list)
    assert streams.query("itag eq 244 and @webm") == [streams.get_by_itag("244")]


def test_query_with_unknown_alias(streams_in_list: list[AudioOrVideoStream]):
    with pytest.raises(QueryError) as exc_info:
        streams = Streams(streams_in_list)
        streams.query("@unknown")
    assert "Unknown alias" in str(exc_info.value)


def test_query_with_function(streams_in_list: list[AudioOrVideoStream]):
    streams = Streams(streams_in_list)
    assert streams.query("best(format eq webm)") == [streams.get_by_itag("271")]


def test_union(streams_in_list: list[AudioOrVideoStream]):
    a, b, *_ = streams_in_list
    assert Streams([a, b]) == Streams([a]) | Streams([b])



import re
import time
from urllib.parse import parse_qs, urlparse

from ytpb.exceptions import BadCommandArgument


def normalize_video_url(video_url_or_id: str) -> str:
    video_id_re = r"(?P<video_id>[\w-]{11})"
    patterns = (
        rf"{video_id_re}$",
        rf"https://www\.(?:youtube\.com/watch\?v=|youtu\.be\/){video_id_re}(?![^&])",
    )

    for pattern in patterns:
        if matched := re.match(pattern, video_url_or_id):
            video_url = build_video_url_with_id(matched.group("video_id"))
            break
    else:
        raise BadCommandArgument(
            "Stream URL or ID not matched. Make sure it opens in a browser."
        )

    return video_url


def extract_parameter_from_url(parameter: str, url: str) -> str:
    url_path_parts = urlparse(url).path.split("/")
    try:
        value_index = url_path_parts.index(parameter)
        value = url_path_parts[value_index + 1]
    except ValueError:
        raise Exception(f"parameter '{parameter}' is not in URL")
    except IndexError:
        raise Exception(f"value of '{parameter}' is not in URL")
    return value


def extract_media_type_from_url(url: str) -> tuple[str, str]:
    media_type = extract_parameter_from_url("mime", url)
    type_name, subtype_name = media_type.split("%2F")
    return type_name, subtype_name


def extract_id_from_base_url(base_url: str) -> str:
    return extract_parameter_from_url("id", base_url)[:11]


def extract_id_from_video_url(video_url: str) -> str:
    parsed = urlparse(video_url)
    try:
        return parse_qs(parsed.query)["v"][0]
    except KeyError:
        return parsed.path.lstrip("/")


def build_video_url_with_id(video_id: str) -> str:
    return f"https://www.youtube.com/watch?v={video_id}"


def build_video_url_from_base_url(base_url: str) -> str:
    video_id = extract_id_from_base_url(base_url)
    return build_video_url_with_id(video_id)


def check_base_url_is_expired(base_url: str) -> bool:
    expires_at = int(extract_parameter_from_url("expire", base_url))
    return time.time() >= expires_at



import re
from dataclasses import dataclass
from enum import auto, StrEnum
from typing import Any

from lxml import etree

from ytpb.exceptions import InfoExtractError


class BroadcastStatus(StrEnum):
    ACTIVE = auto()
    UPCOMING = auto()
    COMPLETED = auto()
    NONE = auto()


@dataclass
class YouTubeVideoInfo:
    """Information about YouTube video."""

    url: str
    title: str
    author: str
    status: BroadcastStatus
    dash_manifest_url: str | None = None


class LeftNotFetched:
    """Represents a value that is intentionally not fetched."""


LEFT_NOT_FETCHED = LeftNotFetched()


def _find_one_or_raise(element: etree.Element, path: str, message: str = "") -> Any:
    """Find first subelement or value which matches the given XPath expression
    or raise an InfoExtractError error."""
    try:
        (result,) = element.xpath(path)
        return result
    except ValueError as exc:
        raise InfoExtractError(message) from exc


def _extract_dash_manifest_url(index_page: str) -> str:
    matched = re.search(r"(?<=dashManifestUrl\":\").*?(?=\")", index_page)
    if not matched:
        raise InfoExtractError("Could not find DASH manifest URL")
    return matched[0]


def extract_video_info(url: str, index_page_text: str) -> YouTubeVideoInfo:
    index_page_element = etree.HTML(index_page_text)

    # Extracting title and author:
    video_object_element = _find_one_or_raise(
        index_page_element,
        './/*[@itemtype="http://schema.org/VideoObject"]',
        "Could not find a http://schema.org/VideoObject element",
    )
    title = _find_one_or_raise(
        video_object_element,
        './meta[@itemprop="name"]/@content',
        "Could not extract a title",
    )
    author = _find_one_or_raise(
        video_object_element,
        './*[@itemtype="http://schema.org/Person"]/link[@itemprop="name"]/@content',
        "Could not extract an author",
    )

    # Check if a video is a live stream or not:
    if (
        broadcast_event_element := index_page_element.find(
            './/*[@itemtype="http://schema.org/BroadcastEvent"]'
        )
    ) is not None:
        is_completed = bool(broadcast_event_element.find('./meta[@itemprop="endDate"]'))
        if is_completed:
            status = BroadcastStatus.COMPLETED
            dash_manifest_url = None
        else:
            status = BroadcastStatus.ACTIVE
            dash_manifest_url = _extract_dash_manifest_url(index_page_text)
    else:
        status = BroadcastStatus.NONE
        dash_manifest_url = None

    return YouTubeVideoInfo(url, title, author, status, dash_manifest_url)



import io
import os
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, patch
from urllib.parse import urljoin

import av
import click
import pytest

from conftest import TEST_DATA_PATH
from freezegun import freeze_time
from PIL import Image

IMAGE_SAVE_SETTINGS = {"format": "jpeg", "quality": 80}


def assert_two_images(
    actual_image_path: Path,
    segment_path: Path,
    expected_frame: int,
    save_settings: dict = IMAGE_SAVE_SETTINGS,
):
    def extract_nth_frame_image(segment_path: Path, n: int) -> Image:
        with av.open(str(segment_path)) as container:
            stream = container.streams.video[0]
            for i, frame in enumerate(container.decode(stream)):
                if i == n:
                    break
        assert frame
        return frame.to_image()

    expected_image = extract_nth_frame_image(segment_path, expected_frame)
    expected_buffer = io.BytesIO()
    expected_image.save(expected_buffer, **save_settings)

    expected_image_saved = Image.open(expected_buffer)
    actual_image = Image.open(actual_image_path)

    assert expected_image_saved.tobytes() == actual_image.tobytes()


@freeze_time("2023-03-26T00:00:00+00:00")
def test_capture_by_sequence_number(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(video_base_url, r"sq/\w+"),
    )

    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "capture",
                "frame",
                "--no-cache",
                "--moment",
                "7959120",
                "-vf",
                "itag eq 244",
                stream_url,
            ],
        )

    assert result.exit_code == 0
    actual_image_path = tmp_path / "Webcam-Zurich-HB_20230325T233354+00.jpg"
    assert os.path.exists(actual_image_path)

    segment_path = TEST_DATA_PATH / "segments" / "7959120.i244.webm"
    assert_two_images(actual_image_path, segment_path, expected_frame=0)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_capture_by_date(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(video_base_url, r"sq/\w+"),
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "capture",
                "frame",
                "--no-cache",
                "--moment",
                "2023-03-25T23:33:55+00",
                "-vf",
                "itag eq 244",
                stream_url,
            ],
            catch_exceptions=False,
            standalone_mode=False,
        )
    assert result.exit_code == 0
    actual_image_path = tmp_path / "Webcam-Zurich-HB_20230325T233355+00.jpg"
    assert os.path.exists(actual_image_path)

    segment_path = TEST_DATA_PATH / "segments" / "7959120.i244.webm"
    assert_two_images(actual_image_path, segment_path, expected_frame=15)


def test_unsupported_output_extension(
    ytpb_cli_invoke: Callable,
    stream_url: str,
) -> None:
    with pytest.raises(click.BadParameter) as exc_info:
        ytpb_cli_invoke(
            [
                "--no-config",
                "capture",
                "frame",
                "--moment",
                "2023-03-25T23:33:55+00",
                "--output",
                "test.unsupported",
                stream_url,
            ],
            catch_exceptions=False,
            standalone_mode=False,
        )
    assert "Format '.unsupported' is not supported" in str(exc_info)


def test_not_provided_output_extension(
    ytpb_cli_invoke: Callable,
    stream_url: str,
) -> None:
    with pytest.raises(click.BadParameter) as exc_info:
        ytpb_cli_invoke(
            [
                "--no-config",
                "capture",
                "frame",
                "--moment",
                "2023-03-25T23:33:55+00",
                "--output",
                "test",
                stream_url,
            ],
            catch_exceptions=False,
            standalone_mode=False,
        )
    assert "Image extension must be provided" in str(exc_info)



import csv
from pathlib import Path
from typing import NamedTuple

import pytest

from conftest import TEST_DATA_PATH

from ytpb.locate import SegmentLocator, SequenceMetadataPair
from ytpb.segment import Segment, SegmentMetadata
from ytpb.types import SegmentSequence


def read_gap_case_fixture_data(path: Path) -> dict:
    class Row(NamedTuple):
        sequence: str
        ingestion_walltime_ms: str
        duration_s: str

    data = {}
    with open(path) as f:
        reader = csv.reader(f, delimiter=",")
        next(reader)
        for fields in reader:
            row = Row(*fields)
            data[int(row.sequence)] = (
                float(row.ingestion_walltime_ms) / 1e6,
                float(row.duration_s),
            )

    return data


@pytest.fixture(scope="class")
def gap_case_fixture_factory(
    request: pytest.FixtureRequest, monkeyclass: pytest.MonkeyPatch
) -> None:
    request.cls.fixture_data = read_gap_case_fixture_data(request.cls.fixture_data_path)

    def _create_segment_metadata(sequence: SegmentSequence) -> SegmentMetadata:
        fake_required_fields = {
            "ingestion_uncertainty": 0,
            "stream_duration": 0,
            "max_dvr_duration": 0,
            "first_frame_time": 0,
            "first_frame_uncertainty": 0,
        }
        return SegmentMetadata(
            sequence_number=sequence,
            ingestion_walltime=request.cls.fixture_data[sequence][0],
            target_duration=2.0,
            **fake_required_fields,
        )

    def mock_download_segment(sequence: SegmentSequence, *args, **kwargs) -> Path:
        return Path(str(sequence))

    def mock_create_segment(sequence_as_path: Path) -> Segment:
        segment = Segment()
        segment.sequence = int(str(sequence_as_path))
        segment.metadata = _create_segment_metadata(segment.sequence)
        return segment

    def mock_get_actual_duration(class_obj: Segment) -> float:
        return request.cls.fixture_data[class_obj.sequence][1]

    def mock_download_segment_and_parse_metadata(
        class_obj: Segment, sequence: SegmentSequence
    ) -> SegmentMetadata:
        return _create_segment_metadata(sequence)

    monkeyclass.setattr("ytpb.locate.download_segment", mock_download_segment)
    monkeyclass.setattr(Segment, "from_file", mock_create_segment)
    monkeyclass.setattr(Segment, "get_actual_duration", mock_get_actual_duration)
    monkeyclass.setattr(
        SequenceMetadataPair,
        "_download_segment_and_parse_metadata",
        mock_download_segment_and_parse_metadata,
    )


@pytest.mark.usefixtures("gap_case_fixture_factory")
class BaseGapCase:
    fixture_data_path: str
    reference_sequence: int

    @classmethod
    def setup_class(cls):
        cls.test_base_url = "https://rr5---sn-5hneknee.googlevideo.com/videoplayback/expire/1679810403/ei/A4sfZK2bNI6HyQWm_ayoBQ/ip/0.0.0.0/id/kHwmzef842g.2/itag/140/source/yt_live_broadcast/requiressl/yes/spc/99c5CWjeTpgWlc2Ht4dNOPt0QhK37Bg/vprv/1/playlist_type/DVR/ratebypass/yes/mime/audio%2Fmp4/live/1/gir/yes/noclen/1/dur/2.000/keepalive/yes/fexp/24007246/sparams/expire,ei,ip,id,itag,source,requiressl,spc,vprv,playlist_type,ratebypass,mime,live,gir,noclen,dur/sig/AOq0QJ8wRQIhAPBxN7yXkU2fPTceJrT_lK4Xw6mT7xWGOjkOrV7Yeh8KAiAq1suuogbWY1qV3dAERpp5I-YUSlVRZQBFlqRpmH-5xQ%3D%3D/mh/XB/mm/44/mn/sn-5hneknee/ms/lva/mt/1679787559/mv/u/mvi/5/pl/24/lsparams/mh,mm,mn,ms,mv,mvi,pl/lsig/AG3C_xAwRQIgfNF-yRJyHRp2Uy4nuoe4-SExdFLMbippDYfUo9FMH0ACIQCEdJrPkdJIilz7Kr-cmdu1Nh3NOwRVdbndOsu-ubkChQ%3D%3D/"

    def setup_method(self):
        self.ssl = SegmentLocator(self.test_base_url, self.reference_sequence)


class TestGapCase1(BaseGapCase):
    fixture_data_path = f"{TEST_DATA_PATH}/gap-cases/gap-case-1-fixture.csv"
    reference_sequence = 7959630

    def test_S1(self):
        assert (7959599, False) == self.ssl.find_sequence_by_time(1679788193.600278)

    def test_E1(self):
        assert (7959599, False) == self.ssl.find_sequence_by_time(
            1679788193.600278, end=True
        )

    @pytest.mark.parametrize("reference", [None, 7959600, 7959601, 7959602])
    def test_S2(self, reference: int | None):
        target_time = 1679788196.600287
        expected_sequence = 7959600
        if reference is None:
            assert (expected_sequence, False) == self.ssl.find_sequence_by_time(
                target_time
            )
        else:
            sl = SegmentLocator(self.test_base_url, reference)
            assert (expected_sequence, False) == sl.find_sequence_by_time(target_time)

    @pytest.mark.parametrize("reference", [None, 7959600, 7959601, 7959602])
    def test_S3(self, reference: int | None):
        """For this case, two segments are possibly valid formally, depending on
        the chosen reference."""
        target_time = 1679788198.599000
        if reference in [None, 7959601]:
            expected_sequence = 7959601
        else:
            expected_sequence = 7959602
        if reference is None:
            assert (expected_sequence, False) == self.ssl.find_sequence_by_time(
                target_time
            )
        else:
            sl = SegmentLocator(self.test_base_url, reference)
            assert (expected_sequence, False) == sl.find_sequence_by_time(target_time)


class TestGapCase2(BaseGapCase):
    fixture_data_path = f"{TEST_DATA_PATH}/gap-cases/gap-case-2-fixture.csv"
    reference_sequence = 7947346

    def test_S1(self):
        assert (7947333, False) == self.ssl.find_sequence_by_time(1679763599.262686)

    def test_S2(self):
        assert (7947334, False) == self.ssl.find_sequence_by_time(1679763601.235894)

    def test_S3(self):
        assert (7947335, True) == self.ssl.find_sequence_by_time(1679763611.742391)

    def test_E3(self):
        assert (7947334, True) == self.ssl.find_sequence_by_time(
            1679763611.742391, end=True
        )

    @pytest.mark.xfail
    def test_S4(self):
        assert (7947335, False) == self.ssl.find_sequence_by_time(1679763626.506922)


class TestGapCase3(BaseGapCase):
    fixture_data_path = f"{TEST_DATA_PATH}/gap-cases/gap-case-3-fixture.csv"
    reference_sequence = 7958122

    def test_S1(self):
        assert (7958102, False) == self.ssl.find_sequence_by_time(1679785199.451019)

    def test_S2(self):
        assert (7958103, False) == self.ssl.find_sequence_by_time(1679785201.449813)

    def test_S3(self):
        assert (7958104, True) == self.ssl.find_sequence_by_time(1679785204.623643)

    def test_E3(self):
        assert (7958103, True) == self.ssl.find_sequence_by_time(
            1679785204.623643, end=True
        )

    def test_S4(self):
        sequence, is_inside_gap = self.ssl.find_sequence_by_time(1679785208.850441)
        assert sequence in (7958104, 7958105)
        assert False == is_inside_gap

    def test_S5(self):
        assert (7958106, False) == self.ssl.find_sequence_by_time(1679785208.903407)



from datetime import datetime, timezone
from pathlib import Path
from textwrap import fill
from typing import Callable

import click
from cloup.constraints import constraint, mutually_exclusive
from PIL import Image

from ytpb.cli import parameters
from ytpb.cli.common import EARLIEST_DATE_TIMEDELTA
from ytpb.utils.path import OUTPUT_PATH_PLACEHOLDER_RE


def validate_output_path(template_context_class) -> Callable[..., Path]:
    def wrapper(ctx: click.Context, param: click.Option, value: Path) -> Path:
        if matched := set(OUTPUT_PATH_PLACEHOLDER_RE.findall(str(value))):
            known_template_vars = template_context_class.__annotations__.keys()
            if not matched.issubset(known_template_vars):
                unknown_vars = matched - set(known_template_vars)
                unknown_vars_string = str(unknown_vars).strip("{}")
                raise click.BadParameter(
                    f"Unknown variable(s) found: {unknown_vars_string}. "
                    f"Option value: '{value}'."
                )
        return value

    return wrapper


def validate_image_output_path(template_context_class) -> Callable[..., Path]:
    def wrapper(ctx: click.Context, param: click.Option, value: Path) -> Path:
        if not value.suffix:
            raise click.BadParameter("Image extension must be provided.")

        extensions = Image.registered_extensions()
        allowed_extensions = {ext for ext, f in extensions.items() if f in Image.SAVE}
        if value.suffix not in allowed_extensions:
            tip = fill(
                "Choose one of: {}".format(", ".join(sorted(allowed_extensions)))
            )
            raise click.BadParameter(
                f"Format '{value.suffix}' is not supported.\n\n{tip}"
            )

        return validate_output_path(template_context_class)(ctx, param, value)

    return wrapper


def validate_start_date_not_too_far(
    ctx: click.Context, param: click.Option, value: datetime | str
) -> datetime | str:
    if isinstance(value, datetime):
        now = datetime.now(timezone.utc).astimezone(value.tzinfo)
        earliest_date = now - EARLIEST_DATE_TIMEDELTA
        if value <= earliest_date:
            raise click.BadParameter(
                "Start date is beyond the limit of 7 days. "
                "The earliest date is {}.".format(
                    earliest_date.isoformat(timespec="minutes")
                )
            )
    return value


interval_option = click.option(
    "-i",
    "--interval",
    metavar="INTERVAL",
    type=parameters.RewindIntervalParamType(),
    help="Time or segment interval.",
    required=True,
)


preview_option = click.option(
    "-p",
    "--preview",
    help="Run in preview mode.",
    is_flag=True,
)


def config_options(f):
    f = click.option(
        "--config",
        "config_path",
        help="Specifies a path to a configuration file.",
        type=click.Path(path_type=Path),
        is_eager=True,
    )(f)

    f = click.option(
        "--no-config",
        help="Do not load any configuration files.",
        default=False,
        is_flag=True,
        is_eager=True,
    )(f)

    return f


def cache_options(f):
    f = click.option(
        "--force-update-cache",
        help="Force to update cache.",
        is_flag=True,
    )(f)

    f = click.option(
        "--no-cache",
        help="Do not use cache.",
        is_flag=True,
    )(f)

    f = constraint(mutually_exclusive, ["no_cache", "force_update_cache"])(f)

    return f


def logging_options(f):
    f = click.option(
        "--debug",
        "debug",
        help="Enable verbose output for debugging.",
        is_flag=True,
        is_eager=True,
    )(f)

    f = click.option(
        "--report",
        help="Dump all output to a file. It implies --debug.",
        is_flag=True,
        is_eager=True,
    )(f)

    return f


yt_dlp_option = click.option(
    "-Y", "--yt-dlp", is_flag=True, help="Use yt-dlp to extract info."
)


no_cleanup_option = click.option(
    "--no-cleanup", is_flag=True, help="Do not clean up temporary files."
)



import json
import os
from datetime import datetime, UTC
from pathlib import Path

from freezegun import freeze_time

from ytpb.cache import read_from_cache, remove_expired_cache_items, write_to_cache


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_remove_expired_items(cache_directory: Path):
    a = cache_directory / "1697012299~a"
    a.touch()
    b = cache_directory / "1697012300~b"
    b.touch()
    c = cache_directory / "1697012301~c"
    c.touch()

    remove_expired_cache_items(cache_directory)
    assert not os.path.exists(a)
    assert not os.path.exists(b)
    assert os.path.exists(c)


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_remove_expired_items_from_empty_cache(cache_directory: Path):
    assert next(cache_directory.iterdir(), None) is None
    remove_expired_cache_items(cache_directory)


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_read_item_from_cache(cache_directory: Path):
    (cache_directory / "1697012301~a").touch()
    (cache_directory / "1697012302~a").write_text('{"f(x)": "x"}')
    assert read_from_cache("a", cache_directory) == {"f(x)": "x"}


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_read_non_existing_item_from_cache(cache_directory: Path):
    assert read_from_cache("n", cache_directory) is None


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_read_expired_item_from_cache(cache_directory: Path):
    a = cache_directory / "1697012299~a"
    a.touch()
    assert read_from_cache("a", cache_directory) is None


@freeze_time(str(datetime.fromtimestamp(1697012300, UTC)))
def test_write_to_cache(cache_directory: Path):
    o = cache_directory / "1697012299~o"
    o.touch()

    a1 = cache_directory / "1697012301~a"
    a1.touch()

    a2 = cache_directory / "1697012302~a"
    item_to_cache = {"f(x)": "x"}
    write_to_cache("a", "1697012302", item_to_cache, cache_directory)
    with open(a2) as f:
        cached_item = json.load(f)

    assert cached_item == item_to_cache

    assert os.path.exists(o)
    assert not os.path.exists(a1)



import re

from conftest import TEST_DATA_PATH
from lxml import etree

from ytpb import mpd


def test_strip_manifest():
    manifest = etree.parse(TEST_DATA_PATH / "manifest-1695928670.mpd")

    stripped_tags = ["<SegmentList", "<SegmentTimeline", "<SegmentURL", r"<S\s"]
    stripped_tags_re = re.compile("|".join(stripped_tags).encode())
    assert stripped_tags_re.search(etree.tostring(manifest))

    output = mpd.strip_manifest(manifest)
    assert output and not stripped_tags_re.search(output)


def test_extract_representations_info(audio_base_url: str, video_base_url: str):
    with open(TEST_DATA_PATH / "manifest-1695928670.mpd") as f:
        results = mpd.extract_representations_info(f.read())

    assert (
        mpd.AudioRepresentationInfo(
            itag="140",
            codecs="mp4a.40.2",
            mime_type="audio/mp4",
            base_url=audio_base_url,
            audio_sampling_rate=44100,
        )
        in results
    )
    assert (
        mpd.VideoRepresentationInfo(
            itag="244",
            codecs="vp9",
            mime_type="video/webm",
            base_url=video_base_url,
            width=854,
            height=480,
            frame_rate=30,
        )
        in results
    )



from dataclasses import asdict

import pytest

from ytpb.segment import Segment


def test_segment_metadata_parsing() -> None:
    content = b"Sequence-Number: 1150301\r\nIngestion-Walltime-Us: 1679329555339525\r\nIngestion-Uncertainty-Us: 85\r\nStream-Duration-Us: 5751479780030\r\nMax-Dvr-Duration-Us: 14400000000\r\nTarget-Duration-Us: 5000000\r\nFirst-Frame-Time-Us: 1679329560650712\r\nFirst-Frame-Uncertainty-Us: 87\r\nEncoding-Alias: L1_Bg\r\n"
    segment = Segment()
    segment.local_path = "test"
    assert asdict(segment.parse_youtube_metadata(content)) == {
        "sequence_number": 1150301,
        "ingestion_walltime": 1679329555.339525,
        "ingestion_uncertainty": 8.5e-5,
        "stream_duration": 5751479.780030,
        "max_dvr_duration": 14400.0,
        "target_duration": 5.0,
        "first_frame_time": 1679329560.650712,
        "first_frame_uncertainty": 8.7e-5,
        "encoding_alias": "L1_Bg",
        "streamable": None,
    }


def test_another_segment_metadata_parsing() -> None:
    content = b"Sequence-Number: 19447394\r\nIngestion-Walltime-Us: 1703082200383647\r\nIngestion-Uncertainty-Us: 80\r\nStream-Duration-Us: 38894190937000\r\nMax-Dvr-Duration-Us: 7200000000\r\nTarget-Duration-Us: 2000000\r\nStreamable: T\r\nFirst-Frame-Time-Us: 1703082201132871\r\nFirst-Frame-Uncertainty-Us: 83\r\nEncoding-Alias: L1_Bg\r\n"
    segment = Segment()
    segment.local_path = "test"
    assert asdict(segment.parse_youtube_metadata(content)) == {
        "sequence_number": 19447394,
        "ingestion_walltime": 1703082200.383647,
        "ingestion_uncertainty": 8.0e-5,
        "stream_duration": 38894190.937000,
        "max_dvr_duration": 7200.0,
        "target_duration": 2.0,
        "first_frame_time": 1703082201.132871,
        "first_frame_uncertainty": 8.3e-5,
        "encoding_alias": "L1_Bg",
        "streamable": "T",
    }


def test_get_actual_duration(audio_segment: Segment) -> None:
    assert pytest.approx(audio_segment.get_actual_duration()) == 1.996916



from collections.abc import Iterator
from datetime import datetime

import av
import structlog
from PIL import Image

from ytpb.locate import SegmentLocator
from ytpb.playback import Playback
from ytpb.segment import Segment
from ytpb.types import SegmentSequence

logger = structlog.get_logger(__name__)


def extract_frame_as_image(
    segment: Segment, target_date: datetime, last_as_fallback: bool = True
) -> Image:
    target_offset = (target_date - segment.ingestion_start_date).total_seconds()
    with av.open(str(segment.local_path)) as container:
        stream = container.streams.video[0]
        target_pts = stream.start_time + target_offset / stream.time_base
        previous_frame: av.VideoFrame = None
        for current_frame in container.decode(stream):
            if current_frame.pts >= target_pts:
                break
            previous_frame = current_frame
        else:
            if last_as_fallback:
                message = "Target date is out of the video"
                logger.debug(f"{message}, use last frame", date=target_date)
            else:
                raise ValueError(message)
    return (previous_frame or current_frame).to_image()


def capture_frames(
    playback: Playback,
    target_dates: list[datetime],
    base_url: str,
    reference_sequence: SegmentSequence,
) -> Iterator[Image, Segment]:
    number_of_targets = len(target_dates)
    previous_sequence = reference_sequence
    for i, target_date in enumerate(target_dates):
        sl = SegmentLocator(
            base_url,
            reference_sequence=previous_sequence,
            temp_directory=playback.get_temp_directory(),
            session=playback.session,
        )
        is_end = i == number_of_targets - 1
        found_sequence, _ = sl.find_sequence_by_time(target_date.timestamp(), is_end)
        previous_sequence = found_sequence

        segment = playback.get_downloaded_segment(found_sequence, base_url)
        image = extract_frame_as_image(segment, target_date)

        yield image, segment



import os
from pathlib import Path
from urllib.parse import urljoin

import responses

from ytpb.download import download_segment


def test_download_segment(
    add_responses_callback_for_segment_urls, audio_base_url: str, tmp_path: Path
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(urljoin(audio_base_url, r"sq/\w+"))

    # When:
    output_path = download_segment(7959120, audio_base_url, tmp_path)

    # Then:
    assert output_path == tmp_path / "7959120.i140.mp4"
    assert os.path.exists(output_path)


def test_download_segment_with_custom_output_filename(
    mocked_responses: responses.RequestsMock, audio_base_url: str, tmp_path: Path
) -> None:
    mocked_responses.get(urljoin(audio_base_url, "sq/0"))

    # When:
    output_path = download_segment(0, audio_base_url, tmp_path, "custom")

    # Then:
    assert output_path == tmp_path / "custom"
    assert os.path.exists(output_path)


def test_download_segment_with_custom_callable_output_filename(
    mocked_responses: responses.RequestsMock, audio_base_url: str, tmp_path: Path
) -> None:
    mocked_responses.get(urljoin(audio_base_url, "sq/0"))

    def custom_output_filename(sequence, base_url):
        assert sequence == 0
        assert base_url == audio_base_url
        return "custom"

    # When:
    output_path = download_segment(0, audio_base_url, tmp_path, custom_output_filename)

    # Then:
    assert output_path == tmp_path / "custom"
    assert os.path.exists(output_path)



from pathlib import Path
from typing import NamedTuple

import structlog
from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)

from ytpb.download import download_segment
from ytpb.merge import merge_segments
from ytpb.playback import Playback, RewindInterval

logger = structlog.get_logger(__name__)


class _ExcerptDownloadResult(NamedTuple):
    exception: Exception | None
    merged_path: Path | None
    audio_segment_paths: list[Path]
    video_segment_paths: list[Path]


def download_excerpt(
    playback: Playback,
    rewind_interval: RewindInterval,
    audio_format_spec: str | None = None,
    video_format_spec: str | None = None,
    output_directory: str | Path | None = None,
    output_stem: str | Path | None = None,
    no_merge: bool = False,
    **merge_kwargs,
) -> _ExcerptDownloadResult:
    sequences_to_download = range(
        rewind_interval.start.sequence, rewind_interval.end.sequence + 1
    )

    download_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(bar_width=28),
        MofNCompleteColumn(),
        TaskProgressColumn(),
        TextColumn("eta"),
        TimeRemainingColumn(),
        console=Console(),
    )

    if audio_format_spec:
        if queried := playback.streams.query(audio_format_spec):
            audio_base_url = queried[0].base_url
            audio_download_task = download_progress.add_task(
                "   - Audio", total=len(sequences_to_download)
            )
        else:
            raise AssertionError

    if video_format_spec:
        if queried := playback.streams.query(video_format_spec):
            video_base_url = queried[0].base_url
            video_download_task = download_progress.add_task(
                "   - Video", total=len(sequences_to_download)
            )
        else:
            raise AssertionError

    segments_output_directory = playback.get_temp_directory() / "segments"
    segments_output_directory.mkdir(parents=True, exist_ok=True)
    download_segment_kwargs = {
        "output_directory": segments_output_directory,
        "session": playback.session,
    }

    downloaded_audio_paths: list[Path] = []
    downloaded_video_paths: list[Path] = []

    with download_progress:
        for sequence in sequences_to_download:
            if audio_format_spec:
                downloaded_path = download_segment(
                    sequence, audio_base_url, **download_segment_kwargs
                )
                downloaded_audio_paths.append(downloaded_path)
                download_progress.advance(audio_download_task)
            if video_format_spec:
                downloaded_path = download_segment(
                    sequence, video_base_url, **download_segment_kwargs
                )
                downloaded_video_paths.append(downloaded_path)
                download_progress.advance(video_download_task)

    if no_merge:
        result = _ExcerptDownloadResult(
            None,
            None,
            downloaded_audio_paths,
            downloaded_video_paths,
        )
    else:
        try:
            merged_path = merge_segments(
                downloaded_audio_paths,
                downloaded_video_paths,
                output_directory=output_directory,
                output_stem=output_stem,
                temp_directory=playback.get_temp_directory(),
                **merge_kwargs,
            )
            result = _ExcerptDownloadResult(
                None, merged_path, downloaded_audio_paths, downloaded_video_paths
            )
        except Exception as e:
            result = _ExcerptDownloadResult(
                e, None, downloaded_audio_paths, downloaded_video_paths
            )

    return result



from ytpb.types import RelativePointInStream


S_TO_MS = 1e3
US_TO_S = 1e-6


def resolve_relativity_in_interval(start, end):
    is_start_relative = isinstance(start, RelativePointInStream)
    is_end_relative = isinstance(end, RelativePointInStream)

    if is_start_relative and is_end_relative:
        raise ValueError("Start and end couldn't be both relative")

    try:
        if is_start_relative:
            start = end - start
        elif is_end_relative:
            end = start + end
    except TypeError:
        pass

    return start, end



from datetime import datetime

from ytpb.types import DateInterval


class TestDateInterval:
    interval = DateInterval(datetime(2023, 12, 10, 1), datetime(2023, 12, 10, 2))

    def test_duration(self):
        assert self.interval.duration == 60 * 60

    def test_substraction_of_intervals(self):
        other = DateInterval(
            datetime(2023, 12, 10, 0, 50), datetime(2023, 12, 10, 2, 10)
        )
        assert self.interval - other == (10 * 60, -10 * 60)
        assert other - self.interval == (-10 * 60, 10 * 60)

    def test_date_inside_interval(self):
        assert datetime(2023, 12, 10, 1, 30) in self.interval
        assert datetime(2023, 12, 10, 1) in self.interval

    def test_interval_is_subinterval(self):
        assert self.interval.is_subinterval(
            DateInterval(datetime(2023, 12, 10, 0, 50), datetime(2023, 12, 10, 2, 10))
        )
        assert not self.interval.is_subinterval(
            DateInterval(datetime(2023, 12, 10, 1, 10), datetime(2023, 12, 10, 1, 50))
        )



from dataclasses import asdict, dataclass
from difflib import unified_diff

import freezegun
import pytest
from freezegun import freeze_time

from helpers import patched_freezgun_astimezone

from ytpb.actions.compose import compose_dynamic_mpd, compose_static_mpd, refresh_mpd
from ytpb.exceptions import YtpbError
from ytpb.playback import Playback
from ytpb.segment import SegmentMetadata
from ytpb.streams import Streams
from ytpb.types import AudioOrVideoStream, AudioStream, VideoStream

freezegun.api.FakeDatetime.astimezone = patched_freezgun_astimezone


@dataclass
class FakeRewindMoment:
    sequence: int


@dataclass
class FakeRewindInterval:
    start: FakeRewindMoment
    end: FakeRewindMoment


@pytest.fixture()
def testing_static_mpd(audio_base_url: str, video_base_url: str) -> str:
    return f"""<?xml version='1.0' encoding='UTF-8'?>
<!--This file is created with ytpb, and expires at 2023-09-28T21:17:50+02:00-->
<MPD xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="urn:mpeg:DASH:schema:MPD:2011" xsi:schemaLocation="urn:mpeg:DASH:schema:MPD:2011 DASH-MPD.xsd" profiles="urn:mpeg:dash:profile:isoff-main:2011" type="static" mediaPresentationDuration="PT66S">
  <ProgramInformation>
    <Title>Webcam Zürich HB</Title>
    <Source>https://www.youtube.com/watch?v=kHwmzef842g</Source>
  </ProgramInformation>
  <Period duration="PT66S">
    <AdaptationSet id="0" mimeType="audio/mp4" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" duration="2000" timescale="1000"/>
      <Representation id="140" codecs="mp4a.40.2" startWithSAP="1" audioSamplingRate="44100">
        <BaseURL>{audio_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
    <AdaptationSet id="1" mimeType="video/webm" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" duration="2000" timescale="1000"/>
      <Representation id="244" codecs="vp9" startWithSAP="1" width="854" height="480" maxPlayoutRate="1" frameRate="30">
        <BaseURL>{video_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
  </Period>
</MPD>
"""


@pytest.fixture()
def testing_dynamic_mpd(audio_base_url: str, video_base_url: str) -> str:
    return f"""<?xml version='1.0' encoding='UTF-8'?>
<!--This file is created with ytpb, and expires at 2023-09-28T21:17:50+02:00-->
<MPD xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="urn:mpeg:DASH:schema:MPD:2011" xsi:schemaLocation="urn:mpeg:DASH:schema:MPD:2011 DASH-MPD.xsd" profiles="urn:mpeg:dash:profile:isoff-live:2011" type="dynamic" availabilityStartTime="2023-03-25T23:33:54.491+00:00">
  <ProgramInformation>
    <Title>Webcam Zürich HB</Title>
    <Source>https://www.youtube.com/watch?v=kHwmzef842g</Source>
  </ProgramInformation>
  <Period start="PT15917833.539S">
    <AdaptationSet id="0" mimeType="audio/mp4" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" timescale="1000">
        <SegmentTimeline>
          <S d="2000" r="-1"/>
        </SegmentTimeline>
      </SegmentTemplate>
      <Representation id="140" codecs="mp4a.40.2" startWithSAP="1" audioSamplingRate="44100">
        <BaseURL>{audio_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
    <AdaptationSet id="1" mimeType="video/webm" subsegmentAlignment="true">
      <SegmentTemplate media="sq/$Number$" startNumber="7959120" timescale="1000">
        <SegmentTimeline>
          <S d="2000" r="-1"/>
        </SegmentTimeline>
      </SegmentTemplate>
      <Representation id="244" codecs="vp9" startWithSAP="1" width="854" height="480" maxPlayoutRate="1" frameRate="30">
        <BaseURL>{video_base_url}</BaseURL>
      </Representation>
    </AdaptationSet>
  </Period>
</MPD>
"""


@freeze_time(tz_offset=2)
def test_compose_static_mpd(
    stream_url: str,
    fake_info_fetcher: "FakeInfoFetcher",
    streams_in_list: list[AudioOrVideoStream],
    testing_static_mpd: str,
):
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()

    streams = Streams(streams_in_list)

    output = compose_static_mpd(
        playback,
        FakeRewindInterval(FakeRewindMoment(7959120), FakeRewindMoment(7959120 + 32)),
        streams.filter(lambda x: x.itag in ("140", "244")),
    )
    assert output == testing_static_mpd


@freeze_time(tz_offset=2)
def test_compose_dynamic_mpd(
    stream_url: str,
    fake_info_fetcher: "FakeInfoFetcher",
    streams_in_list: list[AudioOrVideoStream],
    testing_dynamic_mpd: str,
):
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()

    streams = Streams(streams_in_list)

    rewind_segment_metadata = SegmentMetadata(
        sequence_number=7959120,
        ingestion_walltime=1679787234.491176,
        ingestion_uncertainty=0,
        stream_duration=15917833.539,
        max_dvr_duration=0,
        target_duration=2,
        first_frame_time=1679787235.764305,
        first_frame_uncertainty=0,
    )

    output = compose_dynamic_mpd(
        playback,
        rewind_segment_metadata,
        streams.filter(lambda x: x.itag in ("140", "244")),
    )
    assert output == testing_dynamic_mpd


@freeze_time(tz_offset=2)
def test_refresh_mpd(streams_in_list: list[dict], testing_static_mpd: str):
    selected_streams_list = []
    for stream in streams_in_list:
        if stream.itag in ("140", "244"):
            stream_dict = asdict(stream)
            stream_dict["base_url"] = "https://test/expire/1695929000/"
            if "audio" in stream_dict["mime_type"]:
                updated_stream = AudioStream(**stream_dict)
            else:
                updated_stream = VideoStream(**stream_dict)
            selected_streams_list.append(updated_stream)

    updated_streams = Streams(selected_streams_list)

    output = refresh_mpd(testing_static_mpd, updated_streams)

    actual_diff = list(
        unified_diff(
            testing_static_mpd.splitlines(keepends=True),
            output.splitlines(keepends=True),
            n=0,
        )
    )
    assert actual_diff == [
        "--- \n",
        "+++ \n",
        "@@ -2 +2 @@\n",
        "-<!--This file is created with ytpb, and expires at 2023-09-28T21:17:50+02:00-->\n",
        "+<!--This file is created with ytpb, and expires at 2023-09-28T21:23:20+02:00-->\n",
        "@@ -12 +12 @@\n",
        "-        <BaseURL>https://rr5---sn-25ge7nzr.googlevideo.com/videoplayback/expire/1695928670/ei/_nwVZYXhAqbQvdIPjKmqgAM/ip/0.0.0.0/id/kHwmzef842g.2/itag/140/source/yt_live_broadcast/requiressl/yes/spc/UWF9fy2D4rPPhPMeyQnmxgP0Yhyaohs/vprv/1/playlist_type/DVR/ratebypass/yes/mime/audio%2Fmp4/live/1/gir/yes/noclen/1/dur/2.000/keepalive/yes/fexp/24007246/beids/24350017/sparams/expire,ei,ip,id,itag,source,requiressl,spc,vprv,playlist_type,ratebypass,mime,live,gir,noclen,dur/sig/AOq0QJ8wRAIgANge9FK8aJnP8nDX_HCd9LixBc1iiZueVKgR1eWAi4ACIE5wyoXt2JUnPjHbh6xp8ZJhy1j9ScEgHiBAO_2xH3h9/initcwndbps/623750/mh/XB/mm/44/mn/sn-25ge7nzr/ms/lva/mt/1695906793/mv/m/mvi/5/pl/38/lsparams/initcwndbps,mh,mm,mn,ms,mv,mvi,pl/lsig/AG3C_xAwRQIhAM6lQ9DNT724pGLtqWR01mXgxu_67Ing2nzPBj4ffCT8AiAYnVuWcAosv-DKUGO2bNSq5ptYGJhRCdlYo8E3-6HKOA%3D%3D/</BaseURL>\n",
        "+        <BaseURL>https://test/expire/1695929000/</BaseURL>\n",
        "@@ -18 +18 @@\n",
        "-        <BaseURL>https://rr5---sn-25ge7nzr.googlevideo.com/videoplayback/expire/1695928670/ei/_nwVZYXhAqbQvdIPjKmqgAM/ip/0.0.0.0/id/kHwmzef842g.2/itag/244/source/yt_live_broadcast/requiressl/yes/spc/UWF9fy2D4rPPhPMeyQnmxgP0Yhyaohs/vprv/1/playlist_type/DVR/ratebypass/yes/mime/video%2Fwebm/live/1/gir/yes/noclen/1/dur/2.000/keepalive/yes/fexp/24007246/beids/24350017/sparams/expire,ei,ip,id,itag,source,requiressl,spc,vprv,playlist_type,ratebypass,mime,live,gir,noclen,dur/sig/AOq0QJ8wRgIhAJBYRElUjO7WhY5_gsjtj0aUbXbyb9Z_Yjo7JeecnqrzAiEAkzwV4SYIFponf7BddjJ5hscSZr8hbPBSx09Qffev9AA%3D/initcwndbps/623750/mh/XB/mm/44/mn/sn-25ge7nzr/ms/lva/mt/1695906793/mv/m/mvi/5/pl/38/lsparams/initcwndbps,mh,mm,mn,ms,mv,mvi,pl/lsig/AG3C_xAwRQIhAP_FmY_xO0cSx-hk2oibYFE1AHaCvDHeYyMXXUEuBNeVAiARmaf6MprHE-eEJJx3Ai59WyTOSt8INUUWhA7MSoEO2w%3D%3D/</BaseURL>\n",
        "+        <BaseURL>https://test/expire/1695929000/</BaseURL>\n",
    ]


def test_refresh_mpd_with_mismatched_streams(
    streams_in_list: list[AudioOrVideoStream], testing_static_mpd: str
):
    streams = Streams(streams_in_list)
    with pytest.raises(YtpbError) as exc_info:
        refresh_mpd(testing_static_mpd, streams.filter(lambda x: x.itag == "140"))
    assert str(exc_info.value) == "No stream with itag '244' in the streams"



from __future__ import annotations

from typing import TYPE_CHECKING

import requests

if TYPE_CHECKING:
    from ytpb.info import BroadcastStatus


class YtpbError(Exception):
    """Base Ytpb exception."""


class BroadcastStatusError(YtpbError):
    def __init__(self, message: str, status: BroadcastStatus):
        self.status = status
        super().__init__(message)


class InfoExtractError(YtpbError):
    pass


class BaseUrlExpiredError(YtpbError):
    pass


class CachedItemNotFoundError(YtpbError):
    def __str__(self) -> str:
        return "Unexpired cached item doesn't exist for the video"


class MaxRetryError(YtpbError):
    """Raised when retry limit has been exceeded."""

    def __init__(self, message, response):
        self.response = response
        super().__init__(message)


class FFmpegRunError(YtpbError):
    pass


class QueryError(YtpbError):
    pass


class BadCommandArgument(YtpbError):
    pass


class SequenceLocatingError(YtpbError):
    pass


class SegmentDownloadError(YtpbError, requests.exceptions.HTTPError):
    def __init__(self, message: str, sequence: int):
        self.sequence = sequence
        super().__init__(message)



import shlex
import subprocess
from pathlib import Path
from typing import Any

import structlog

from ytpb.exceptions import FFmpegRunError

logger = structlog.get_logger(__name__)


def run_ffmpeg(
    args: str | list[str], **subprocess_kwargs: Any
) -> subprocess.CompletedProcess:
    command = ["ffmpeg", "-v", "error", "-hide_banner", "-y"]

    if isinstance(args, str):
        command.extend(shlex.split(args))
    else:
        command.extend(args)

    cp = subprocess.run(command, **subprocess_kwargs)

    try:
        cp.check_returncode()
    except subprocess.CalledProcessError as e:
        logger.error("There was a problem running FFmpeg command.")
        raise FFmpegRunError from e

    return cp


def run_ffprobe(
    input_path: Path, args: str | list[str], **subprocess_kwargs: Any
) -> subprocess.CompletedProcess:
    command = ["ffprobe", "-v", "0", "-hide_banner"]

    if isinstance(args, str):
        command.extend(shlex.split(args))
    else:
        command.extend(args)
    command.append(input_path)

    cp = subprocess.run(command, **subprocess_kwargs)
    return cp


def ffprobe_show_entries(input_path: Path, entries_to_show: str) -> str:
    command_args = f"-show_entries {entries_to_show} -of default=nw=1:nk=1"
    cp = run_ffprobe(input_path, command_args, capture_output=True, check=True)
    output = cp.stdout.decode().rstrip("\n")
    return output


def ffmpeg_stream_copy(input_path: Path, output_path: Path):
    subprocess_kwargs = {"capture_output": True, "check": True}
    run_ffmpeg(f"-i {input_path} -c copy {output_path}", **subprocess_kwargs)



import enum
import math
import re
import string
import warnings
from dataclasses import dataclass
from datetime import datetime, time, timedelta
from typing import Any, Literal


@dataclass
class ISO8601DateStyleParameters:
    format: Literal["basic", "extended"] = "basic"
    precision: Literal["reduced", "complete"] = "complete"
    offset_format: Literal["hh", "hhmm"] = "hh"
    fractional_component: Literal["sec"] | None = None
    fraction_delimiter: Literal[".", ","] = "."
    use_z_for_utc: bool = False


def ensure_date_aware(date: datetime) -> datetime:
    """Ensure a date is timezone aware."""
    if date.tzinfo is None:
        local_tzinfo = date.astimezone().tzinfo
        return date.replace(tzinfo=local_tzinfo)
    else:
        return date


def format_iso_datetime(
    date: datetime,
    style: ISO8601DateStyleParameters | None = None,
) -> str:
    """Format `datetime` objects after ISO 8601. Supports complete and reduced
    date and time representations in basic and extended formats.

    Complete:
    - [year]["-"][month]["-"][day]["T"][hour][min][sec]
    """

    if style is None:
        style = ISO8601DateStyleParameters()

    plain_offset = date.strftime("%z")
    if plain_offset == "":
        raise ValueError("datetime object should be timezone aware")

    match style.precision:
        case "reduced":
            if date.microsecond != 0:
                # output_time_format = "%H:%M:%S.%f"
                output_time_format = "%H:%M:%S"
            elif date.second != 0:
                output_time_format = "%H:%M:%S"
            elif date.minute != 0:
                output_time_format = "%H:%M"
            else:
                output_time_format = "%H"
            output_format = f"%Y-%m-%dT{output_time_format}"
        case "complete":
            output_format = "%Y-%m-%dT%H:%M:%S"
            # if date.microsecond != 0:
            #    output_format += ".%f"
        case _:
            raise ValueError(
                "'precision' value should be either 'reduced' or 'complete'"
            )

    match style.format:
        case "basic":
            output_format = output_format.replace("-", "").replace(":", "")
            full_offset = plain_offset
        case "extended":
            full_offset = f"{plain_offset[:3]}:{plain_offset[-2:]}"
        case _:
            raise ValueError("'format' value should be either 'basic' or 'extended'")

    match style.offset_format:
        case "hh":
            offset = full_offset[:3]
        case "hhmm":
            offset = full_offset
        case _:
            raise ValueError("'offset_format' value should be either 'hh' or 'hhmm'")

    if style.use_z_for_utc and plain_offset == "+0000":
        offset = "Z"

    date_string = date.strftime(output_format)
    if date.microsecond != 0:
        # output = f"{date_string[:-3]}{offset}"
        output = f"{date_string}{offset}"
    else:
        output = f"{date_string}{offset}"

    return output


class DurationFormatPattern(enum.Enum):
    NUMERIC = "%H:%M%:%S"
    ISO8601 = "PT[%-HH][%-MM][%-SS]"
    IN_SENTENCE = "[%-H h ][%-M m ][%-S s]"


def format_duration(duration: timedelta, pattern: DurationFormatPattern) -> str:
    total_seconds = math.floor(duration.total_seconds() + 0.5)
    total_minutes, ss = divmod(total_seconds, 60)
    hh, mm = divmod(total_minutes, 60)

    output = pattern.value
    time_object = time(hh, mm, ss)

    optional_parts_re = r"\[(?P<part>(?P<fmt>%-?[HMS]).*?)\]"
    for matched in re.finditer(optional_parts_re, pattern.value):
        if int(time_object.strftime(matched.group("fmt"))) == 0:
            output = output.replace(matched.group(0), "")
        else:
            formatted_part = time_object.strftime(matched.group("part"))
            output = output.replace(matched.group(0), formatted_part)

    output = time_object.strftime(output).rstrip(" ")

    return output


def format_timedelta(duration: timedelta, use_ms_precision: bool = False) -> str:
    duration_total_seconds = abs(duration.total_seconds())

    if use_ms_precision:
        total_seconds = int(duration_total_seconds)
    else:
        total_seconds = math.floor(duration_total_seconds + 0.5)
    total_minutes, ss = divmod(total_seconds, 60)
    hh, mm = divmod(total_minutes, 60)

    sign = "+" if duration.total_seconds() >= 0 else "-"
    if hh == 0:
        output = f"{sign}{mm:02}:{ss:02}"
    else:
        output = f"{sign}{hh:02}:{mm:02}:{ss:02}"

    if use_ms_precision:
        rounded_milliseconds = round(abs(duration.microseconds / 1e3))
        output += f".{rounded_milliseconds:03d}"

    return output


def round_date(date: datetime) -> datetime:
    rounded = date.replace(second=0, microsecond=0)
    if date.microsecond < 1e6 / 2:
        rounded = rounded + timedelta(seconds=date.second)
    else:
        rounded = rounded + timedelta(seconds=date.second + 1)
    return rounded


def express_timedelta_in_words(delta: timedelta) -> str:
    if delta.seconds > 60:
        total_seconds = math.floor(delta.seconds / 60 + 0.5) * 60
    else:
        total_seconds = delta.seconds

    total_minutes, ss = divmod(total_seconds, 60)
    hh, mm = divmod(total_minutes, 60)

    if (hh == 0 and mm >= 45) or (hh == 1 and mm < 15):
        output = "an hour"
    elif hh >= 1:
        x = hh + math.floor(mm / 60 / 0.5 + 0.5) * 0.5
        output = f"{x:g} hours"
    elif mm == 1 and ss < 30:
        output = "a minute"
    elif mm >= 1:
        x = mm + math.floor(ss / 60 + 0.5)
        output = f"{x:g} minutes"
    elif mm == 0 and ss > 10:
        output = "less than a minute"
    else:
        output = "a moment"

    return output


def build_style_parameters_from_spec(
    style_spec: str,
) -> ISO8601DateStyleParameters | None:
    if not style_spec:
        return None

    style_parameters = {}
    input_styles = set([x.strip() for x in style_spec.split(",")])

    if "z" in input_styles:
        style_parameters["use_z_for_utc"] = True
        input_styles.remove("z")

    conflicting_styles_map: dict[str, str] = {}
    parameters_to_check = ["format", "precision", "offset_format"]
    for field_name in parameters_to_check:
        field = ISO8601DateStyleParameters.__dataclass_fields__[field_name]
        conflicting_styles_map[field_name] = field.type.__args__

    for input_style in input_styles:
        for parameter, parameter_styles in conflicting_styles_map.items():
            if input_style in parameter_styles:
                if already_set_style := style_parameters.get(parameter, None):
                    conflicted_styles = sorted([already_set_style, input_style])
                    raise ValueError(
                        "Mutually exclusive styles provided: "
                        f"'{conflicted_styles[0]}' and '{conflicted_styles[1]}'"
                    )
                else:
                    style_parameters[parameter] = input_style
                    break

    if unknown_styles := input_styles ^ set(style_parameters.values()):
        warnings.warn(f"Ignoring unknown style(s): {', '.join(sorted(unknown_styles))}")

    return ISO8601DateStyleParameters(**style_parameters)


class ISO8601DateFormatter(string.Formatter):
    """A variant of `string.Formatter` that format dates in ISO 8601
    format.

    This extends the format specification for `datetime` objects with
    the following, ISO-8601 related, styles: 'basic', 'extended', 'reduced',
    'complete', 'hh', 'hhmm', and 'z'.
    """

    def format_field(self, value: Any, format_spec: str) -> str:
        if isinstance(value, datetime):
            if not format_spec:
                return format_iso_datetime(value)
            style_parameters = build_style_parameters_from_spec(format_spec)
            output = format_iso_datetime(value, style_parameters)  # type: ignore
        else:
            output = super().format_field(value, format_spec)

        return output



import requests

from ytpb.exceptions import YtpbError
from ytpb.types import AudioOrVideoStream, SegmentSequence, VideoStream


def get_priority_reference_stream(streams: list[AudioOrVideoStream]) -> str:
    for stream in streams:
        if isinstance(stream, VideoStream):
            break
    return stream


def request_reference_sequence(
    base_url: str, session: requests.Session | None = None
) -> SegmentSequence:
    session = session or requests.Session()
    response = session.head(base_url, allow_redirects=True)
    try:
        return int(response.headers["X-Head-Seqnum"])
    except KeyError:
        raise YtpbError("'X-Head-Seqnum' header value is missing")
